/*******************************************************************************

        @file AbstractReader.d
        
        Copyright (C) 2004 Kris Bell
        
        This software is provided 'as-is', without any express or implied
        warranty. In no event will the authors be held liable for damages
        of any kind arising from the use of this software.
        
        Permission is hereby granted to anyone to use this software for any 
        purpose, including commercial applications, and to alter it and/or 
        redistribute it freely, subject to the following restrictions:
        
        1. The origin of this software must not be misrepresented; you must 
           not claim that you wrote the original software. If you use this 
           software in a product, an acknowledgment within documentation of 
           said product would be appreciated but is not required.

        2. Altered source versions must be plainly marked as such, and must 
           not be misrepresented as being the original software.

        3. This notice may not be removed or altered from any distribution
           of the source.


                        ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


        @version        Initial version, October 2004      
        @author         Kris


*******************************************************************************/

module mango.io.AbstractReader;

private import  mango.io.Exception,
                mango.io.ArrayAllocator;

public  import  mango.io.model.IReader,
                mango.io.model.IBuffer,
                mango.io.model.IConduit;


/*******************************************************************************

        Reader base-class. Each reader operates upon an IBuffer, which is
        provided at construction time. Said buffer is intended to remain 
        consistent over the reader lifetime.

	Readers support both a C++ iostream type syntax, along with a get()
	syntax. Operations may be chained back-to-back.

	All readers support the full set of native data types, plus a
	selection of array types. The latter can be configured to produce
	either a copy (.dup) of the buffer content, or a slice. See class
	SimpleAllocator, BufferAllocator and SliceAllocator for more on 
        this topic.

	The code below illustrates basic operation upon a memory buffer:
	
	@code
        Buffer buf = new Buffer (256);

        // map same buffer into both reader and writer
        IReader r = new Reader(buf);
        IWriter w = new Writer(buf);

        int i = 10;
        long j = 20;
        double d = 3.14159;
        char[] c = "fred";

        // write data types out
        w << c << i << j << d;

        // read them back again
        r >> c >> i >> j >> d;

        // reset
        buf.clear();

        // same thing again, but using get() syntax instead
        w.put(c).put(i).put(j).put(d);
        r.get(c).get(i).get(j).get(d);

	@endcode

	Readers may also be used with any class implementing the IReadable
	interface. See PickleReader for an example of how this can be put
        to good use.
	
*******************************************************************************/

class AbstractReader : SimpleAllocator, IReader
{       
	/***********************************************************************

	***********************************************************************/

	struct NumericDecoder
	{
		BufferConverter         int1,
					int8,
					int16,
					int32,
					int64,
					int8u,
					int16u,
					int32u,
					int64u,
					float32,
					float64,
					float80;
	}

	/***********************************************************************

	***********************************************************************/

	struct StringDecoder
	{
		BufferConverter         char8,
					char16,
					char32;
	}

        // a couple of pre-constructed exceptions 
        private static EofException     eof;
        private static IOException      udf;

        // the buffer associated with this reader. Note that this
        // must not change for the lifetime of the reader, since
        // it is assumed to be immutable elsewhere (subclass may
        // keep its own copy).
        protected IBuffer               buffer;         

        // decoders for string requests
        protected StringDecoder         string;

        // decoders for numeric requests
        protected NumericDecoder        numeric;

        // memory-manager for array requests
        private IArrayAllocator         memory;

        // size for client-specified allocations
        private uint                    pushed = uint.max;

        /***********************************************************************
        
                Return the name of this reader

        ***********************************************************************/

        abstract char[] toString ();

        /***********************************************************************
        
                Construct a couple of static exception instances.

        ***********************************************************************/

        static this ()
        {
                eof = new EofException ("premature end-of-input");
                udf = new IOException  ("input buffer underflow");
        }

        /***********************************************************************
        
                Return the buffer associated with this reader

        ***********************************************************************/

        this (IBuffer buffer)
        {
                super (this);
                this.memory = this;
                this.buffer = buffer;
        }

        /***********************************************************************
        
                Return the buffer associated with this reader

        ***********************************************************************/

        final IBuffer getBuffer ()
        {
                return buffer;
        }

        /***********************************************************************
        
                Return the allocator associated with this reader.See 
                ArrayAllocator for more information.

        ***********************************************************************/

        final IArrayAllocator getAllocator ()
        {
                return memory;
        }

        /***********************************************************************
        
                Set the allocator to use for array management. Arrays are
                always allocated by the IReader. That is, you cannot read
                data into an array slice (for example). Instead, a number
                of IArrayAllocator classes are available to manage memory
                allocation when reading array content. 

                By default, an IReader will allocate each array from the 
                heap. You can change that behavior by calling this method
                with an IArrayAllocator of choice. For instance, there 
                is a BufferAllocator which will slice an array directly 
                from the buffer where possible. Also available is the 
                record-oriented SliceAllocator, which slices memory from 
                within a pre-allocated heap area, and should be reset by
                the client code after each record has been read (to avoid 
                unnecessary growth).

                See ArrayAllocator for more information.

        ***********************************************************************/

        final void setAllocator (IArrayAllocator memory) 
        {
                if (! memory)
                      memory = this;
                this.memory = memory;
        }

        /***********************************************************************
        
                Set the configured IStringDecoder. These are intended to
                be used as a conversion mechanism between various character
                representations. They are also expected to be used for the
                process of applying character encodings.
        
                See IStringDecoder.

        ***********************************************************************/

        final void setStringDecoder (IStringDecoder s) 
        {
                string.char8  = &s.char8;
                string.char16 = &s.char16;
                string.char32 = &s.char32;
        }

        /***********************************************************************
        
                Push the size (in bytes) to use for the next array-read.
                By default, array sizes are read from the input stream, 
                so this is the means by which one may specify the size
                where the stream is not formatted in such a manner.

                @code
                        char[]  x;
                        int     size;
                        IReader reader;

                        reader.push(size).get(x);
                @endcode
                
        ***********************************************************************/

        final IReader push (uint size)
        {
                if (pushed == uint.max)
                    pushed = size;
                else
                   throw new IOException ("invalid IReader push-sequence");
                return this;
        }

        /***********************************************************************
        
                Read and return an integer from the input stream. This can
                be used to extract the length of a subsequent array.

        ***********************************************************************/

        final uint length ()
        {
                uint len = pushed;

                if (len != uint.max)
                    pushed = uint.max;
                else
                   get (len);
                return len;
        }

        /***********************************************************************
        
                Wait for something to arrive in the buffer. This may stall
                the current thread forever, although usage of SocketConduit 
                will take advantage of the timeout facilities provided there.

        ***********************************************************************/

        final void wait ()
        {
                buffer.get (1, false);
        }

        /***********************************************************************
        
		Extract a readable class from the current read-position
		
        ***********************************************************************/

        final IReader get (IReadable x) 
        {
                assert (x);
                x.read (this); 
                return this;
        }

        /***********************************************************************

		Extract a boolean value from the current read-position	
	        
        ***********************************************************************/

        IReader get (inout bool x)
        {
                numeric.int1 (&x, x.sizeof);
                return this;
        }

        /***********************************************************************

		Extract an unsigned byte value from the current read-position	
	        		
        ***********************************************************************/

        IReader get (inout ubyte x) 
        {
                numeric.int8u (&x, x.sizeof);
                return this;
        }

        /***********************************************************************
        
		Extract a byte value from the current read-position
		
        ***********************************************************************/

        IReader get (inout byte x)
        {
                numeric.int8 (&x, x.sizeof);
                return this;
        }

        /***********************************************************************
        
		Extract an unsigned short value from the current read-position
		
        ***********************************************************************/

        IReader get (inout ushort x)
        {
                numeric.int16u (&x, x.sizeof);
                return this;
        }

        /***********************************************************************
        
		Extract a short value from the current read-position
		
        ***********************************************************************/

        IReader get (inout short x)
        {
                numeric.int16 (&x, x.sizeof);
                return this;
        }

        /***********************************************************************
        
		Extract a unsigned int value from the current read-position
		
        ***********************************************************************/

        IReader get (inout uint x)
        {
                numeric.int32u (&x, x.sizeof);
                return this;
        }

        /***********************************************************************
        
		Extract an int value from the current read-position
		
        ***********************************************************************/

        IReader get (inout int x)
        {
                numeric.int32 (&x, x.sizeof);
                return this;
        }

        /***********************************************************************
        
		Extract an unsigned long value from the current read-position
		
        ***********************************************************************/

        IReader get (inout ulong x)
        {
                numeric.int64u (&x, x.sizeof);
                return this;
        }

        /***********************************************************************
        
		Extract a long value from the current read-position
		
        ***********************************************************************/

        IReader get (inout long x)
        {
                numeric.int64 (&x, x.sizeof);
                return this;
        }

        /***********************************************************************
        
		Extract a float value from the current read-position
		
        ***********************************************************************/

        IReader get (inout float x)
        {
                numeric.float32 (&x, x.sizeof);
                return this;
        }

        /***********************************************************************
        
		Extract a double value from the current read-position
		
        ***********************************************************************/

        IReader get (inout double x)
        {
                numeric.float64 (&x, x.sizeof);
                return this;
        }

        /***********************************************************************
        
		Extract a real value from the current read-position
		
        ***********************************************************************/

        IReader get (inout real x)
        {
                numeric.float80 (&x, x.sizeof);
                return this;
        }

        /***********************************************************************
        
		Extract a char value from the current read-position
		
        ***********************************************************************/

        IReader get (inout char x)
        {
                string.char8 (&x, x.sizeof);
                return this;
        }

        /***********************************************************************
        
		Extract a wide char value from the current read-position
		
        ***********************************************************************/

        IReader get (inout wchar x)
        {
                string.char16 (&x, x.sizeof);
                return this;
        }

        /***********************************************************************
        
		Extract a double char value from the current read-position
		
        ***********************************************************************/

        IReader get (inout dchar x)
        {
                string.char32 (&x, x.sizeof);
                return this;
        }

        /***********************************************************************
        
        ***********************************************************************/

        final IReader get (inout int[] x)
        {
                memory.allocate (cast(void[]*) &x, string.char8);
                return this;
        }

        /***********************************************************************
        
        ***********************************************************************/

        IReader get (inout char[] x)
        {
                memory.allocate (cast(void[]*) &x, string.char8);
                return this;
        }

        /***********************************************************************
        
        ***********************************************************************/

        IReader get (inout wchar[] x)
        {
                memory.allocate (cast(void[]*) &x, string.char16);
                return this;
        }

        /***********************************************************************
        
        ***********************************************************************/

        IReader get (inout dchar[] x)
        {
                memory.allocate (cast(void[]*) &x, string.char32);
                return this;
        }

        

        /**********************************************************************/
        /***************** opShr() is just an alias for get() *****************/
        /**********************************************************************/


        version (UseShiftOperators)
        {
                /***************************************************************

                ***************************************************************/

                IReader opShr (inout bit x)
                {
                        return get (x);
                }

                /***************************************************************

                ***************************************************************/

                IReader opShr (inout ubyte x)
                {
                        return get (x);
                }

                /***************************************************************

                ***************************************************************/

                IReader opShr (inout byte x)
                {
                        return get (x);
                }

                /***************************************************************

                ***************************************************************/

                IReader opShr (inout ushort x)
                {
                        return get (x);
                }

                /***************************************************************

                ***************************************************************/

                IReader opShr (inout short x)
                {
                        return get (x);
                }

                /***************************************************************

                ***************************************************************/

                IReader opShr (inout uint x)
                {
                        return get (x);
                }

                /***************************************************************

                ***************************************************************/

                IReader opShr (inout int x)
                {
                        return get (x);
                }

                /***************************************************************

                ***************************************************************/

                IReader opShr (inout ulong x)
                {
                        return get (x);
                }

                /***************************************************************

                ***************************************************************/

                IReader opShr (inout long x)
                {
                        return get (x);
                }

                /***************************************************************

                ***************************************************************/

                IReader opShr (inout float x)
                {
                        return get (x);
                }

                /***************************************************************

                ***************************************************************/

                IReader opShr (inout double x)
                {
                        return get (x);
                }

                /***************************************************************

                ***************************************************************/

                IReader opShr (inout real x)
                {
                        return get (x);
                }

                /***************************************************************

                ***************************************************************/

                IReader opShr (IReadable x) 
                {
                        return get (x);
                }

                /***************************************************************

                ***************************************************************/

                IReader opShr (inout char x)
                {
                        return get (x);
                }

                /***************************************************************

                ***************************************************************/

                IReader opShr (inout wchar x)
                {
                        return get (x);
                }

                /***************************************************************

                ***************************************************************/

                IReader opShr (inout dchar x)
                {
                        return get (x);
                }

                /***************************************************************

                ***************************************************************/

                IReader opShr (inout char[] x)
                {       
                        return get (x);
                }

                /***************************************************************

                ***************************************************************/

                IReader opShr (inout wchar[] x)
                {
                        return get (x);
                }

                /***************************************************************

                ***************************************************************/

                IReader opShr (inout dchar[] x)
                {
                        return get (x);
                }
        }
}


/+
        /***********************************************************************

        ***********************************************************************/

        template Get (T, alias decoder)
        {
                IReader get (inout T x)
                {
                        decoder.get (&x, x.sizeof);
                        return this;
                }
        }

        template GetArray (T, alias decoder)
        {
                IReader get (inout T x)
                {
                        memory.allocate (cast(void[]*) &x, decoder);
                        return this;
                }
        }

        mixin Get!(byte, numeric.int8);
        //mixin GetArray!(byte[], numeric.int8);
+/
