/*******************************************************************************

        @file AbstractWriter.d
        

        Copyright (C) 2004 Kris Bell
        
        This software is provided 'as-is', without any express or implied
        warranty. In no event will the authors be held liable for damages
        of any kind arising from the use of this software.
        
        Permission is hereby granted to anyone to use this software for any 
        purpose, including commercial applications, and to alter it and/or 
        redistribute it freely, subject to the following restrictions:
        
        1. The origin of this software must not be misrepresented; you must 
           not claim that you wrote the original software. If you use this 
           software in a product, an acknowledgment within documentation of 
           said product would be appreciated but is not required.

        2. Altered source versions must be plainly marked as such, and must 
           not be misrepresented as being the original software.

        3. This notice may not be removed or altered from any distribution
           of the source.


                        ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


        @version        Initial version, October 2004      
        @author         Kris

*******************************************************************************/

module mango.io.AbstractWriter;

private import  mango.io.Exception;

public  import  mango.io.model.IWriter,
                mango.io.model.IBuffer,
                mango.io.model.IConduit;

/*******************************************************************************

        Writer base-class. Writers provide the means to append formatted 
        data to an IBuffer, and expose a convenient method of handling a
        variety of data types. In addition to writing native types such
        as integer and char[], writers also process any class which has
        implemented the IWritable interface (one method).

	Writers support both a C++ iostream type syntax, along with a put()
	syntax. Operations may be chained back-to-back.

	All writers support the full set of native data types, plus a
	selection of array types. The latter can be configured to produce
	either a copy (.dup) of the buffer content, or a slice. See class
	SimpleAllocator, BufferAllocator and SliceAllocator for more on 
        this topic.

	The code below illustrates basic operation upon a memory buffer:
	
	@code
        Buffer buf = new Buffer (256);

        // map same buffer into both reader and writer
        IReader r = new Reader(buf);
        IWriter w = new Writer(buf);

        int i = 10;
        long j = 20;
        double d = 3.14159;
        char[] c = "fred";

        // write data types out
        w << c << i << j << d;

        // read them back again
        r >> c >> i >> j >> d;

        // reset
        buf.clear();

        // same thing again, but using put() syntax instead
        w.put(c).put(i).put(j).put(d);
        r.get(c).get(i).get(j).get(d);

	@endcode

	Writers may also be used with any class implementing the IWritable
	interface. See PickleReader for an example of how this can be put
	to good use.
	
*******************************************************************************/

class AbstractWriter : IWriter
{       
        /***********************************************************************

        ***********************************************************************/

	struct NumericEncoder
	{
		BufferConverter         int1,
					int8,
					int16,
					int32,
					int64,
					int8u,
					int16u,
					int32u,
					int64u,
					float32,
					float64,
					float80;
	}

	/***********************************************************************

	***********************************************************************/

	struct StringEncoder
	{
		BufferConverter         char8,
					char16,
					char32;
	}

        // public newline adaptor
        static INewlineWriter           newline;                

        // a couple of pre-constructed exceptions 
        protected static IOException    ovf;
        protected static EofException   eof;

        protected IBuffer               buffer;

        protected StringEncoder         string;

        protected NumericEncoder        numeric;

        private bool                    prefixArray = true;

        /***********************************************************************
        
                Return the name of this writer

        ***********************************************************************/

        abstract char[] toString ();

        /***********************************************************************
        
                Construct some static exception instances, and create the
                public 'newline' instance.

        ***********************************************************************/

        static this ()
        {
                newline = new NewlineWriter;

                eof = new EofException ("eof during write");
                ovf = new IOException  ("output buffer too small");
        }

        /***********************************************************************
        
                Construct a Writer upon the provided IBuffer. All formatted
                output will be appended to this buffer.

        ***********************************************************************/

        this (IBuffer buffer)
        {
                this.buffer = buffer;
        }
     
        /***********************************************************************
        
                Return the associated buffer

        ***********************************************************************/

        final IBuffer getBuffer ()
        {     
                return buffer;
        }

        /***********************************************************************
        
                Set the configured IStringEncoder. These are intended to
                be used as a conversion mechanism between various character
                representations. They are also expected to be used for the
                process of applying character encodings.

                See IStringEncoder.

        ***********************************************************************/

        final void setStringEncoder (IStringEncoder s) 
        {
                string.char8  = &s.char8;
                string.char16 = &s.char16;
                string.char32 = &s.char32;
        }

        /***********************************************************************
        
                Flush the output of this writer. Returns false if the 
                operation failed, true otherwise.

        ***********************************************************************/

        IWriter flush ()
        {  
                buffer.flush ();
                return this;
        }

        /***********************************************************************
        
                Output a newline. Do this indirectly so that it can be 
                intercepted by subclasses.

        ***********************************************************************/

        IWriter cr ()
        {
                return put (newline);
        }

        /***********************************************************************

        ***********************************************************************/

        void enableArrayPrefix (bool on)
        {
                prefixArray = on;
        }

        /***********************************************************************
        
                Write a class to the current buffer-position
                
        ***********************************************************************/

        private final uint length (uint len)
        {
                if (prefixArray)
                    put (len);
                return len;
        }

        /***********************************************************************
        
                Write a class to the current buffer-position
                
        ***********************************************************************/

        IWriter put (IWritable x) 
        {
                assert (x);
                x.write (this); 
                return this;
        }

        /***********************************************************************
        
                Write a char value to the current buffer-position
                
        ***********************************************************************/

        IWriter put (char x)
        {
                string.char8 (&x, x.sizeof);
                return this;
        }

        /***********************************************************************
        
                Write a wide char value to the current buffer-position
                
        ***********************************************************************/

        IWriter put (wchar x)
        {
                string.char16 (&x, x.sizeof);
                return this;
        }

        /***********************************************************************
        
                Write a double char value to the current buffer-position
                
        ***********************************************************************/

        IWriter put (dchar x)
        {
                string.char32 (&x, x.sizeof);
                return this;
        }

        /***********************************************************************
        
                Write a char array to the current buffer-position
                
        ***********************************************************************/

        IWriter put (char[] x) 
        {
                string.char8 (x, length (x.length * char.sizeof));
                return this;
        }

        /***********************************************************************
        
                Write a char array to the current buffer-position
                
        ***********************************************************************/

        IWriter putw (wchar[] x) 
        {
                string.char16 (x, length (x.length * wchar.sizeof));
                return this;
        }

        /***********************************************************************
        
                Write a char array to the current buffer-position
                
        ***********************************************************************/

        IWriter putd (dchar[] x)
        {
                string.char32 (x, length (x.length * dchar.sizeof));
                return this;
        }

        /***********************************************************************
        
                Write a boolean value to the current buffer-position    
                
        ***********************************************************************/

        IWriter put (bool x)
        {
                numeric.int1 (&x, x.sizeof);
                return this;
        }

        /***********************************************************************
        
                Write an unsigned byte value to the current buffer-position     
                                
        ***********************************************************************/

        IWriter put (ubyte x)
        {
                numeric.int8u (&x, x.sizeof);
                return this;
        }

        /***********************************************************************
        
                Write a byte value to the current buffer-position
                
        ***********************************************************************/

        IWriter put (byte x)
        {
                numeric.int8 (&x, x.sizeof);
                return this;
        }

        /***********************************************************************
        
                Write an unsigned short value to the current buffer-position
                
        ***********************************************************************/

        IWriter put (ushort x)
        {
                numeric.int16u (&x, x.sizeof);
                return this;
        }

        /***********************************************************************
        
                Write a short value to the current buffer-position
                
        ***********************************************************************/

        IWriter put (short x)
        {
                numeric.int16 (&x, x.sizeof);
                return this;
        }

        /***********************************************************************
        
                Write a unsigned int value to the current buffer-position
                
        ***********************************************************************/

        IWriter put (uint x)
        {
                numeric.int32u (&x, x.sizeof);
                return this;
        }

        /***********************************************************************
        
                Write an int value to the current buffer-position
                
        ***********************************************************************/

        IWriter put (int x)
        {
                numeric.int32 (&x, x.sizeof);
                return this;
        }

        /***********************************************************************
        
                Write an unsigned long value to the current buffer-position
                
        ***********************************************************************/

        IWriter put (ulong x)
        {
                numeric.int64u (&x, x.sizeof);
                return this;
        }

        /***********************************************************************
        
                Write a long value to the current buffer-position
                
        ***********************************************************************/

        IWriter put (long x)
        {
                numeric.int64 (&x, x.sizeof);
                return this;
        }

        /***********************************************************************
        
                Write a float value to the current buffer-position
                
        ***********************************************************************/

        IWriter put (float x)
        {
                numeric.float32 (&x, x.sizeof);
                return this;
        }

        /***********************************************************************
        
                Write a double value to the current buffer-position
                
        ***********************************************************************/

        IWriter put (double x)
        {
                numeric.float64 (&x, x.sizeof);
                return this;
        }

        /***********************************************************************
        
                Write a real value to the current buffer-position
                
        ***********************************************************************/

        IWriter put (real x)
        {
                numeric.float80 (&x, x.sizeof);
                return this;
        }


        /**********************************************************************/
        /*************** opShl() is just an alias for put *********************/
        /**********************************************************************/


        version (UseShiftOperators)
        {      
                /***************************************************************

                ***************************************************************/

                IWriter opShl (bit x)
                {
                        return put (x);
                }

                /***************************************************************

                ***************************************************************/

                IWriter opShl (ubyte x)
                {
                        return put (x);
                }

                /***************************************************************

                ***************************************************************/

                IWriter opShl (byte x)
                {
                        return put (x);
                }

                /***************************************************************

                ***************************************************************/

                IWriter opShl (ushort x)
                {
                        return put (x);
                }

                /***************************************************************

                ***************************************************************/

                IWriter opShl (short x)
                {
                        return put (x);
                }

                /***************************************************************

                ***************************************************************/

                IWriter opShl (uint x)
                {
                        return put (x);
                }

                /***************************************************************

                ***************************************************************/

                IWriter opShl (int x)
                {
                        return put (x);
                }

                /***************************************************************

                ***************************************************************/

                IWriter opShl (ulong x)
                {
                        return put (x);
                }

                /***************************************************************

                ***************************************************************/

                IWriter opShl (long x)
                {
                        return put (x);
                }

                /***************************************************************

                ***************************************************************/

                IWriter opShl (float x)
                {
                        return put (x);
                }

                /***************************************************************

                ***************************************************************/

                IWriter opShl (double x)
                {
                        return put (x);
                }

                /***************************************************************

                ***************************************************************/

                IWriter opShl (real x)
                {
                        return put (x);
                }

                /***************************************************************

                ***************************************************************/

                IWriter opShl (IWritable x) 
                {
                        return put (x);
                }

                /***************************************************************

                ***************************************************************/

                IWriter opShl (char x)
                {
                        return put (x);
                }

                /***************************************************************

                ***************************************************************/

                IWriter opShl (wchar x)
                {
                        return put (x);
                }

                /***************************************************************

                ***************************************************************/

                IWriter opShl (dchar x)
                {
                        return put (x);
                }

                /***************************************************************

                ***************************************************************/

                IWriter opShl (char[] x)
                {
                        return put (x);
                }

                /***************************************************************

                ***************************************************************/

                IWriter opShl (wchar[] x) 
                {
                        return putw (x);
                }

                /***************************************************************

                ***************************************************************/

                IWriter opShl (dchar[] x)
                {
                        return putd (x);
                }
        }
}


/*******************************************************************************

        A class to handle newline output. One might reasonable expect to 
        emit a char[] for newlines; FileSystem.NewlineString for example.
        Turns out that it's much more efficient to intercept line-breaks
        when they're implemented in a more formal manner (such as this).

        For example, ColumnWriter() and TextWriter() both must intercept 
        newline output so they can adjust formatting appropriately. It is 
        much more efficient for such writers to intercept the IWritable 
        put() method instead of scanning each char[] for the various \\n 
        combinations.
        
        Please use the INewlineWriter interface for emitting newlines.

*******************************************************************************/

private import mango.io.FileSystem;

class NewlineWriter : INewlineWriter
{
        private char[]  fmt;

        /***********************************************************************

                Construct a default newline, using the char[] defined 
                by FileSystem.NewlineString
        
        ***********************************************************************/

        this ()
        {
                this (FileSystem.NewlineString);
        }

        /***********************************************************************
        
                Construct a newline using the provided character array

        ***********************************************************************/

        this (char[] fmt)
        {
                this.fmt = fmt;
        }

        /***********************************************************************
        
                Write this newline through the provided writer. This makes
                NewlineWriter IWritable compatible.

        ***********************************************************************/

        void write (IWriter w)
        {
                w.put (fmt);
        }     
}
