/*******************************************************************************

        @file CompositeReader.d
        
        Copyright (C) 2004 Kris Bell
        
        This software is provided 'as-is', without any express or implied
        warranty. In no event will the authors be held liable for damages
        of any kind arising from the use of this software.
        
        Permission is hereby granted to anyone to use this software for any 
        purpose, including commercial applications, and to alter it and/or 
        redistribute it freely, subject to the following restrictions:
        
        1. The origin of this software must not be misrepresented; you must 
           not claim that you wrote the original software. If you use this 
           software in a product, an acknowledgment within documentation of 
           said product would be appreciated but is not required.

        2. Altered source versions must be plainly marked as such, and must 
           not be misrepresented as being the original software.

        3. This notice may not be removed or altered from any distribution
           of the source.


                        ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


        @version        Initial version, March 2004      
        @author         Kris


*******************************************************************************/

module mango.io.CompositeReader;

private import  mango.io.Exception;

public  import  mango.io.model.IBuffer,
                mango.io.model.IReader;


/*******************************************************************************

        CompositeReader 'brackets' the reading of a data set.  

        The approach leverages the exisiting IReadable interface which, 
        when implemented, exposes a class that reads a related set of 
        data. For instance:
        
        @code
                class Foo : IReadable
                {
                        int     x,
                                y;
                        char[]  text;

                        void read (IReader reader)
                        {
                                reader.get(x).get(y).get(text);
                        }
                }
        @endcode

        In the above example, we set up a class that implements IReadable
        via the read() method. The class may read whatever it chooses,
        and may also invoke the read() method of other IReadable classes.
        Thus, the initial IReadable has control over a potential readable
        cluster, all of which is bracketed within the CompositeReader. 
        
        This CompositeReader ensure it has enough space in the buffer to 
        read an entire data-set, such that buffer slices mapped into user
        space will all be valid. This does not apply where mapping is not
        enabled (which is the default setting for all readers).

        A word on buffer-mapping: There are a number of situations where
        a programmer prefers to avoid the overhead of memory allocation &
        subsequent copying of data. Other times, the liveness of content
        is more important. The former is attractive in high-performance 
        systems, whereas the latter is typical of most casual IO.

        Most programs won't reap any benefit from buffer-mapping, but may
        run afoul of liveness concerns. Therefore, mapping is disabled on
        all readers by default. That is, all array types are allocated &
        copied as they are read. Change this via Reader.setMapped().

        Note that buffer-mapping applies only to array[] data types. All
        integral types are copied directly into local program space.

	See CompositeWriter for additional examples.

*******************************************************************************/

class CompositeReader
{
        private int     width;
        private IReader reader;
           
        /***********************************************************************
        
                Construct a composite with the given reader. The reader 
                should be configured for non-mapped input (the default
                state for Readers).

        ***********************************************************************/

        this (IReader reader)
        in {
           assert (reader);
           }
        body
        {
                this.reader = reader;
        }

        /***********************************************************************
        
                Construct a composite with the given reader, and the maximum 
                width of an input data-set. The width is used to ensure that
                an entire data-set will be present in the buffer at one time. 
                The reader is configured for mapped input.

                Use the other constructor(s) for non-mapped input.

        ***********************************************************************/

        this (IReader reader, int width)
        in {
           assert (reader);
           assert (width > 0 && width <= reader.getBuffer().getCapacity());
           }
        body
        {
                this.reader = reader;
                this.width = width;

                // enable mapped arrays
                reader.setMapped (true);
        }

        /***********************************************************************
                
                Read a data-set, after ensuring there's enough buffer space
                for mapping the entire content (where necessary).

        ***********************************************************************/

        void get (IReadable target) 
        in {
           assert (target);
           }
        body
        { 
                IBuffer buffer = reader.getBuffer();

                // ensure we have space for the entire data-set, such that all
                // buffer slices are guaranteed to be valid.
                if (reader.isMapped() && buffer.readable() < width)
                    buffer.compress();

                target.read (reader);
        }
}
