/*******************************************************************************

        @file CompositeWriter.d
        
        Copyright (C) 2004 Kris Bell
        
        This software is provided 'as-is', without any express or implied
        warranty. In no event will the authors be held liable for damages
        of any kind arising from the use of this software.
        
        Permission is hereby granted to anyone to use this software for any 
        purpose, including commercial applications, and to alter it and/or 
        redistribute it freely, subject to the following restrictions:
        
        1. The origin of this software must not be misrepresented; you must 
           not claim that you wrote the original software. If you use this 
           software in a product, an acknowledgment within documentation of 
           said product would be appreciated but is not required.

        2. Altered source versions must be plainly marked as such, and must 
           not be misrepresented as being the original software.

        3. This notice may not be removed or altered from any distribution
           of the source.


                        ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


        @version        Initial version, March 2004      
        @author         Kris


*******************************************************************************/

module mango.io.CompositeWriter;

private import  mango.io.Exception;

public  import  mango.io.model.IBuffer,
                mango.io.model.IWriter;


/*******************************************************************************

        CompositeWriter brackets the writing of a data set, providing a
	basic framework that can be extended to support commit semantics.  

        The approach leverages the exisiting IWritable interface which, 
        when implemented, exposes a class that writes a related set of 
        data. For instance:

        @code
                class Foo : IWritable
                {
                        int     x,
                                y;
                        char[]  text;

                        void write (IWriter writer)
                        {
                                writer.put(x).put(y).put(text);
                        }
                }
        @endcode

        In the above example, we set up a class that implements IWritable
        (via the write() method). The class may write whatever it chooses,
        and may also invoke the write() method of other IWritable classes.
        Thus, the initial IWritable has control over a potential writable
        cluster, all of which is bracketed within the CompositeWriter. 
	CompositeWriter could be extended to support commit and rollback 
	semantics with little effort. 
        
        This particular CompositeWriter simply flushes the output buffer 
	after a data-set has been written to it (option). In general usage
        this will be measurably more efficient than using a FlushBuffer.  

	Here's an expanded example of how this is intended to operate:

	@code
        // define a serializable class (via interfaces)
        class Wumpus : IReadable, IWritable
        {
                private int     x = 11,
                                y = 112,
                                z = 1024;

                void write(IWriter writer)
                {
                        writer.put(x).put(y).put(z);
                }

                void read(IReader reader)
                {
                        reader.get(x).get(y).get(z);
                }
        }

        // construct a Wumpus
        Wumpus wumpus = new Wumpus();

        // open a file for IO
        FileConduit fc = new FileConduit ("random.bin", FileStyle.ReadWriteCreate);

        // construct composite reader & writer upon the file, with binary IO
        CompositeWriter cw = new CompositeWriter (new Writer(fc));
        CompositeReader cr = new CompositeReader (new Reader(fc));

        // write the Wumpus (and flush it)
        cw.put (wumpus);

        // rewind to file start
        fc.seek (0);

        // read Wumpus back again
        cr.get (wumpus);
	@endcode

*******************************************************************************/

class CompositeWriter
{
        private IWriter writer;
               
        /***********************************************************************
        
                Construct a CompositeWriter with the provided IWriter. 

        ***********************************************************************/

        this (IWriter writer)
        in {
           assert (writer);
           }
        body
        {
                this.writer = writer;
        }

        /***********************************************************************
                
                Write the source 
        
        ***********************************************************************/

        CompositeWriter put (IWritable source) 
        in {
           assert (source);
           }
        body
        { 
                // write the contents
                source.write (writer);
                return this;
        }

        /***********************************************************************
                
                Flush the output
        
        ***********************************************************************/

        void flush () 
        {
                // flush data-set to conduit
                writer.flush ();
        }
}

