/*******************************************************************************

        @file DisplayWriter.d
        
        Copyright (C) 2004 Kris Bell
        
        This software is provided 'as-is', without any express or implied
        warranty. In no event will the authors be held liable for damages
        of any kind arising from the use of this software.
        
        Permission is hereby granted to anyone to use this software for any 
        purpose, including commercial applications, and to alter it and/or 
        redistribute it freely, subject to the following restrictions:
        
        1. The origin of this software must not be misrepresented; you must 
           not claim that you wrote the original software. If you use this 
           software in a product, an acknowledgment within documentation of 
           said product would be appreciated but is not required.

        2. Altered source versions must be plainly marked as such, and must 
           not be misrepresented as being the original software.

        3. This notice may not be removed or altered from any distribution
           of the source.


                        ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


        @version        Initial version, March 2004      
        @author         Kris


*******************************************************************************/

module mango.io.DisplayWriter;

private import  std.c.stdio;

public  import  mango.io.Writer;

/*******************************************************************************

        Format output suitable for presentation. There is no corresponding
        reader, since there are no explicit delimiters emitted to isolate
        each formatted item. One can use TextReader to read input from the 
        console (or from some other input buffer).

*******************************************************************************/

class DisplayWriter : Writer
{
        // temporary buffer space for encoding
        private char[40] encode;

        // number of floating-point digits to write 
        private uint digits = 2;

        version (Win32)
                 alias _vsnprintf vsnprintf;

        /***********************************************************************
        
                Construct a DisplayWriter upon the specified IBuffer

        ***********************************************************************/

        this (IBuffer buffer)
        {
                super (buffer);
                
                // disable array length output (for strings and so on)
                enableArrayPrefix (false);

                numeric.int1 = &int1;
                numeric.int8 = &int8;
                numeric.int8u = &int8u;
                numeric.int16 = &int16;
                numeric.int16u = &int16u;
                numeric.int32 = &int32;
                numeric.int32u = &int32u;
                numeric.int64 = &int64;
                numeric.int64u = &int64u;

                numeric.float32 = &float32;
                numeric.float64 = &float64;
                numeric.float80 = &float80;
        }
     
        /***********************************************************************
        
                Construct a DisplayWriter upon the specified IConduit

        ***********************************************************************/

        this (IConduit conduit)
        {
                this (conduit.createBuffer);
        }

        /***********************************************************************
        
        ***********************************************************************/
        
        override char[] toString()
        {
                return "display writer";
        }

        /***********************************************************************
        
                Set the number of floating point digits

        ***********************************************************************/

        void setPrecision (uint digits)
        {
                this.digits = digits;
        }

        /***********************************************************************
        
                Convert to a string and send back through the framework
                such that it may be intercepted by a subclass.

        ***********************************************************************/

        final void format (char[] fmt, void* src)
        {
                put (encode[0..vsnprintf (encode, encode.length, fmt, src)]);
        }

        /***********************************************************************
        
        ***********************************************************************/

        final void int8f (ubyte* src, uint count, char[] fmt)
        {     
                while (count)
                      {
                      uint i = *src;
                      format (fmt, &i);

                      ++src;
                      --count;
                      }
        }

        /***********************************************************************
        
        ***********************************************************************/

        final void int16f (ushort* src, uint count, char[] fmt)
        {     
                while (count)
                      {
                      uint i = *src;
                      format (fmt, &i);

                      ++src;
                      count -= 2;
                      }
        }

        /***********************************************************************
        
        ***********************************************************************/

        final void format (void* src, uint count, uint width, char[] fmt)
        {     
                while (count)
                      {
                      format (fmt, src);

                      src += width;
                      count -= width;
                      }
        }

        /***********************************************************************
        
        ***********************************************************************/

        final void int1 (void* src, uint count)
        {     
                while (count)
                      {
                      if (*cast(ubyte *) src)
                          string.char8 (cast(char*) "true", 4);
                      else
                         string.char8 (cast(char*) "false", 5);

                      ++src;
                      --count;
                      }
        }

        /***********************************************************************
        
        ***********************************************************************/

        final void int8 (void* src, uint count)
        {     
                int8f (cast(ubyte*) src, count, "%d"); 
        }

        /***********************************************************************
        
        ***********************************************************************/

        final void int8u (void* src, uint count)
        {     
                int8f (cast(ubyte*) src, count, "%u"); 
        }

        /***********************************************************************
        
        ***********************************************************************/

        final void int16 (void* src, uint count)
        {     
                int16f (cast(ushort*) src, count, "%d"); 
        }

        /***********************************************************************
        
        ***********************************************************************/

        final void int16u (void* src, uint count)
        {     
                int16f (cast(ushort*) src, count, "%u"); 
        }

        /***********************************************************************
        
        ***********************************************************************/

        final void int32 (void* src, uint count)
        {     
                format (src, count, uint.sizeof, "%d"); 
        }

        /***********************************************************************
        
        ***********************************************************************/

        final void int32u (void* src, uint count)
        {     
                format (src, count, uint.sizeof, "%u"); 
        }

        /***********************************************************************
        
        ***********************************************************************/

        final void int64 (void* src, uint count)
        {     
                format (src, count, long.sizeof, "%lld"); 
        }

        /***********************************************************************
        
        ***********************************************************************/

        final void int64u (void* src, uint count)
        {     
                format (src, count, long.sizeof, "%llu"); 
        }

        /***********************************************************************
        
        ***********************************************************************/

        final void float32 (void* src, uint count)
        {     
                format (src, count, float.sizeof, "%.2f"); 
        }

        /***********************************************************************
        
        ***********************************************************************/

        final void float64 (void* src, uint count)
        {     
                format (src, count, double.sizeof, "%.2lf"); 
        }

        /***********************************************************************
        
        ***********************************************************************/

        final void float80 (void* src, uint count)
        {     
                format (src, count, real.sizeof, "%.2Lf"); 
        }
}



