/*******************************************************************************

        @file IReader.d
        
        Copyright (C) 2004 Kris Bell
        
        This software is provided 'as-is', without any express or implied
        warranty. In no event will the authors be held liable for damages
        of any kind arising from the use of this software.
        
        Permission is hereby granted to anyone to use this software for any 
        purpose, including commercial applications, and to alter it and/or 
        redistribute it freely, subject to the following restrictions:
        
        1. The origin of this software must not be misrepresented; you must 
           not claim that you wrote the original software. If you use this 
           software in a product, an acknowledgment within documentation of 
           said product would be appreciated but is not required.

        2. Altered source versions must be plainly marked as such, and must 
           not be misrepresented as being the original software.

        3. This notice may not be removed or altered from any distribution
           of the source.


                        ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


        @version        Initial version, March 2004      
        @author         Kris


*******************************************************************************/

module mango.io.model.IReader;

public import mango.io.model.IBuffer;

/*******************************************************************************

        Any class implementing IReadable becomes part of the Reader framework
        
*******************************************************************************/

interface IReadable
{
        void read (IReader r);
}

/*******************************************************************************

*******************************************************************************/

interface IArrayAllocator
{
        void reset ();

        void allocate (void[]* x, BufferConverter convert);
}

/*******************************************************************************

        Use instances of IStringDecoder to convert between character 'sizes'
        and to unwrap the various character encodings. These may throw an
        exception if they don't support the requested type. For example, an
        exception would/should be thrown where a Unicode-based IStringDecoder 
        is requested to provide a char[] instead of a wchar[].

        An appropriate IStringDecoder should be attached to each IReader, and
        thus be available for subsequent use. A raw binary implementation is 
        attached by default (no decoding at all).

        Note that the 'count' argument is the number of bytes available in
        the 'dst' buffer. The decoders are expected to fill exactly that 
        amount, or throw an exception.

*******************************************************************************/

interface IStringDecoder
{
        void char8  (void *dst, uint count);
        void char16 (void *dst, uint count);
        void char32 (void *dst, uint count);
}

/*******************************************************************************

        All reader instances should implement this interface.

*******************************************************************************/

interface IReader
{
        /***********************************************************************
        
                These are the basic reader methods

        ***********************************************************************/

        IReader get (inout bool x);
        IReader get (inout byte x);
        IReader get (inout ubyte x);
        IReader get (inout short x);
        IReader get (inout ushort x);
        IReader get (inout int x);
        IReader get (inout uint x);
        IReader get (inout long x);
        IReader get (inout ulong x);
        IReader get (inout float x);
        IReader get (inout double x);
        IReader get (inout real x);

        IReader get (inout char x);
        IReader get (inout wchar x);
        IReader get (inout dchar x);

        IReader get (inout char[] x);
        IReader get (inout wchar[] x);
        IReader get (inout dchar[] x);

        // these are currently disabled until a future release
        //IWriter get (inout bool[] x);
        //IWriter get (inout byte[] x);
        //IWriter get (inout short[] x);
        //IWriter get (inout int[] x);
        //IWriter get (inout long[] x);
        //IWriter get (inout ubyte[] x);
        //IWriter get (inout ushort[] x);
        //IWriter get (inout uint[] x);
        //IWriter get (inout ulong[] x);
        //IWriter get (inout float[] x);
        //IWriter get (inout double[] x);
        //IWriter get (inout real[] x);

        /***********************************************************************
        
                This is the mechanism used for binding arbitrary classes 
                to the IO system. If a class implements IReadable, it can
                be used as a target for IReader get() operations. That is, 
                implementing IReadable is intended to transform any class 
                into an IReader adaptor for the content held therein.

        ***********************************************************************/

        IReader get (IReadable x);

        /***********************************************************************
        
                Push the size (in bytes) to use for the next array-read.
                By default, array sizes are read from the input stream, 
                so this is the means by which one may specify the size
                where the stream is not formatted in such a manner.

                @code
                        char[]  x;
                        int     size;
                        IReader reader;

                        reader.push(size).get(x);
                @endcode

        ***********************************************************************/

        IReader push (uint size);

        /***********************************************************************
                
                Pause the current thread until some content arrives in
                the associated input buffer. This may stall forever.

        ***********************************************************************/

        void wait ();

        /***********************************************************************
        
                Return the size (byte count) of the subsequent data stream. 
                This is typically read from the stream as a number, but may 
                be provided by the client code via the push() method.

        ***********************************************************************/

        uint length ();

        /***********************************************************************
        
                Return the buffer associated with this reader

        ***********************************************************************/

        IBuffer getBuffer ();

        /***********************************************************************
        
                Set the allocator to use for array management. Arrays are
                always allocated by the IReader. That is, you cannot read
                data into an array slice (for example). Instead, a number
                of IArrayAllocator classes are available to manage memory
                allocation when reading array content. 

                By default, an IReader will allocate each array from the 
                heap. You can change that behavior by calling this method
                with an IArrayAllocator of choice. For instance, there 
                is a BufferAllocator which will slice an array directly 
                from the buffer where possible. Also available is the 
                record-oriented SliceAllocator, which slices memory from 
                within a pre-allocated heap area, and should be reset by
                the client code after each record has been read (to avoid 
                unnecessary growth).

                See ArrayAllocator for more information.

        ***********************************************************************/

        void setAllocator (IArrayAllocator memory); 

        /***********************************************************************
        
                Set the configured IStringDecoder. These are intended to
                be used as a conversion mechanism between various character
                representations. They are also expected to be used for the
                process of applying character encodings.

        ***********************************************************************/

        void setStringDecoder (IStringDecoder s);

        /***********************************************************************
        
                And these are used for the >> syntax

        ***********************************************************************/

        version (UseShiftOperators)
        {
                IReader opShr (IReadable s);
                IReader opShr (inout bool x);
                IReader opShr (inout ubyte x);
                IReader opShr (inout byte x);
                IReader opShr (inout ushort x);
                IReader opShr (inout short x);
                IReader opShr (inout uint x);
                IReader opShr (inout int x);
                IReader opShr (inout ulong x);
                IReader opShr (inout long x);
                IReader opShr (inout float x);
                IReader opShr (inout double x);
                IReader opShr (inout real x);

                IReader opShr (inout char x);
                IReader opShr (inout wchar x);
                IReader opShr (inout dchar x);

                IReader opShr (inout char[] x);
                IReader opShr (inout wchar[] x);
                IReader opShr (inout dchar[] x);
        }
}
