/*******************************************************************************

        @file IWriter.d
        
        Copyright (C) 2004 Kris Bell
        
        This software is provided 'as-is', without any express or implied
        warranty. In no event will the authors be held liable for damages
        of any kind arising from the use of this software.
        
        Permission is hereby granted to anyone to use this software for any 
        purpose, including commercial applications, and to alter it and/or 
        redistribute it freely, subject to the following restrictions:
        
        1. The origin of this software must not be misrepresented; you must 
           not claim that you wrote the original software. If you use this 
           software in a product, an acknowledgment within documentation of 
           said product would be appreciated but is not required.

        2. Altered source versions must be plainly marked as such, and must 
           not be misrepresented as being the original software.

        3. This notice may not be removed or altered from any distribution
           of the source.


                        ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


        @version        Initial version, March 2004      
        @author         Kris


*******************************************************************************/

module mango.io.model.IWriter;

public import mango.io.model.IBuffer;

/*******************************************************************************

        Interface to make any class compatible with any IWriter.

*******************************************************************************/

interface IWritable
{
        void write (IWriter w);
}

/*******************************************************************************

        Make a signature for IWritable classes to use when they wish to 
        avoid processing by writers such as TokenWriter.

*******************************************************************************/

interface IPhantomWriter : IWritable
{
}

/*******************************************************************************

        Make a signature for Newline classes

*******************************************************************************/

interface INewlineWriter : IPhantomWriter
{
}

/*******************************************************************************

        Use instances of IStringEncoder to convert between character 'sizes'
        and to apply the various character encodings. These may throw an
        exception if they don't support the requested type. For example, an
        exception would/should be thrown where a ASCII-based IStringEncoder 
        is requested to write a dchar[] instead of a char[].

        An appropriate IStringEncoder should be attached to each IWriter, and
        thus be available for subsequent use. A raw binary implementation is
        attached by default (no encoding).

        Note that the 'count' argument is the number of bytes provided in
        the 'src' buffer. The decoders are expected to convert exactly that 
        amount, or throw an exception.

*******************************************************************************/

interface IStringEncoder
{
        void char8  (void *src, uint count);
        void char16 (void *src, uint count);
        void char32 (void *src, uint count);
}

/*******************************************************************************

        IWriter interface. IWriter provides the means to append formatted 
        data to an IBuffer, and exposes a convenient method of handling a
        variety of data types. In addition to writing native types such
        as integer and char[], writers also process any class which has
        implemented the IWritable interface (one method).

*******************************************************************************/

interface IWriter
{
        /***********************************************************************
        
                These are the basic writer methods

        ***********************************************************************/

        IWriter put (bool x);
        IWriter put (ubyte x);
        IWriter put (byte x);
        IWriter put (ushort x);
        IWriter put (short x);
        IWriter put (uint x);
        IWriter put (int x);
        IWriter put (ulong x);
        IWriter put (long x);
        IWriter put (float x);
        IWriter put (double x);
        IWriter put (real x);

        IWriter put (char x);
        IWriter put (wchar x);
        IWriter put (dchar x);

        IWriter put (char[] x);

        // these will have to wait until walter implements string prefixes
        //IWriter put (wchar[] x);
        //IWriter put (dchar[] x);

        // these are currently disabled until a future release
        //IWriter put (bool[] x);
        //IWriter put (byte[] x);
        //IWriter put (short[] x);
        //IWriter put (int[] x);
        //IWriter put (long[] x);
        //IWriter put (ubyte[] x);
        //IWriter put (ushort[] x);
        //IWriter put (uint[] x);
        //IWriter put (ulong[] x);
        //IWriter put (float[] x);
        //IWriter put (double[] x);
        //IWriter put (real[] x);

        IWriter put (IWritable x);

        /***********************************************************************
        
                Set the configured IStringEncoder. These are intended to
                be used as a conversion mechanism between various character
                representations. They are also expected to be used for the
                process of applying character encodings.

        ***********************************************************************/

        void setStringEncoder (IStringEncoder s);

        /***********************************************************************
        
                And these are used for the << syntax. Some are currently 
                unavailable due to the compiler being confused over its
                implicit type conversions ...

        ***********************************************************************/

        version (UseShiftOperators)
        {
                IWriter opShl (bool x);
                IWriter opShl (ubyte x);
                IWriter opShl (byte x);
                IWriter opShl (ushort x);
                IWriter opShl (short x);
                IWriter opShl (uint x);
                IWriter opShl (int x);
                IWriter opShl (ulong x);
                IWriter opShl (long x);
                IWriter opShl (float x);
                IWriter opShl (double x);
                IWriter opShl (real x);

                IWriter opShl (char x);
                IWriter opShl (wchar x);
                IWriter opShl (dchar x);

                IWriter opShl (char[] x);

                // these will have to wait until walter implements string prefixes
                //IWriter opShl (wchar[] x);
                //IWriter opShl (dchar[] x);

                IWriter opShl (IWritable x);
        }

        /***********************************************************************

                Output a newline. Do this indirectly so that it can be 
                intercepted by subclasses.
        
        ***********************************************************************/

        IWriter cr ();

        /***********************************************************************
        
                Flush the output of this writer. Throws an IOException 
                if the operation fails.

        ***********************************************************************/

        IWriter flush ();

        /***********************************************************************
        
                Return the associated buffer

        ***********************************************************************/

        IBuffer getBuffer ();
}

