/******************************************************************************

        @mainpage The Mango Tree

        Mango is a collection of D packages with an orientation toward
        server-side programming. These packages may be used together 
        or in isolation and, in many cases, can be just as applicable 
        to client-side development. Mango is targeted for Win32 and 
        linux platforms.

        There are currently seven packages under the Mango Tree:

        @li mango.io: a high-performance buffered IO package. Primary 
            functionality includes memory Buffers, external Conduits 
            (files, sockets, etc), file-system manipulation, several 
            flavors of Readers and Writers (formatted data), TextFormat 
            (printf) and TextLayout support, class serialization, 
            tokenization, and a whole lot more. Socket support is via 
            the gracious provision of Christopher Miller's socket module. 
            See FilePath, FileProxy, FileConduit, FileSystem, SocketConduit, 
            ServerSocket, MulticastSocket, SocketListener, Token, Stdio, 
            Reader, Writer, IPickle and PickleRegistry.

        @li mango.http.client: An HTTP client for conversing with a remote
            HTTP server. Supports headers, query parameters, cookies, 
            timeout etc. Tokenizer classes in mango.io are handy for 
            parsing returned content. See HttpClient.

        @li mango.http.server: an HTTP server framework is provided for 
            implementing a variety of high performance servers based 
            on the HTTP protocol. This is not a fully-fledged v1.1 
            implementation. That is, it doesn't currently support 
            'fields' or 'keep alive'. However, it happily processes 
            each request without allocating memory, making it unusually 
            efficient. This is an example of how D array-slicing can 
            seriously improve runtime efficiency. An open-source HTTPS 
            library should be fairly easy to bolt in.

        @li mango.servlet: A servlet-style engine that sits upon 
            mango.server. Includes most of the things that Java servlet 
            programmers would expect, and attempts to improve in certain 
            key areas such as the IO system. See Servlet and ServletProvider.

        @li mango.cache: Some simple caching mechanisms. Includes MRU 
            caching via a queue, and level-two caching (to disk) via 
            class-serialization. Now based upon a derivative of Doug Lea's
            concurrent hashmap. The mango.cluster package derives from
            mango.cache, so it's really easy to switch between a local
            cache implementation and a clustered version. See ICache, 
            IMutableCache, PlainCache, QueuedCache and VirtualCache.

        @li mango.log: A clone of the wildly popular Log4J package, for 
            logging and tracing of runtime behaviour. See
            <A HREF="http://logging.apache.org/log4j/docs/documentation.html">this page</A>
            for an introduction to Log4J. Mango.log has a dependency upon 
            mango.io for file access (PropertyConfigurator, FileAppender) 
            and for one of the console appenders (StdioAppender). Other 
            than that it is completely standalone, and it would be simple 
            to replace/augment those appenders with alternatives. Mango.log
            can generate remote log messages, and has an HTML Admin Console
            which dynamically adjusts the log settings of an executing program.
            See Logger, BasicConfigurator and PropertyConfigurator.

        @li mango.cluster: A clustering package for servers and other
            network aware applications. Currently supports clustered
            caching and queuing, where the former is limited to a fixed 
            MRU number of entries per node, and the latter is limited to an
            upper memory waterline. While the underlying QOS (quality of 
            service) substrate is setup to be plugable, the QOS
            currently provided is stream/multicast based. Note that this
            package is not a substitute for a persistence layer; instead
            it is a tool for enabling horizontal scaling within a system 
            that has been designed explicitly from the outset to scale.
            On the other hand, it can significantly enhance the throughput 
            of any persistence layer by offloading a large percentage of
            read operations. See CacheInvalidatee, CacheInvalidator, 
            NetworkCombo, NetworkAlert, NetworkCache and NetworkQueue.

        Package dependencies are as follows (from left to right):

	@code
        mango.io           => phobos   (limited)
        mango.log          => mango.io (limited)
        mango.cache        => mango.io 
        mango.http.server  => mango.io
        mango.http.client  => mango.io, mango.server
        mango.cluster      => mango.io, mango.cache, mango.log
        mango.servlet      => mango.io, mango.http.server, mango.cache, mango.log
 	@endcode

        You might take a look at the unittest.d source to get a general 
        feel for functionality, and browse the various pdf files 
        <A HREF="http://svn.dsource.org/svn/projects/mango/trunk/doc/">here</A>       
        along with the class-hierarchy or alphabetical-list linked at the 
        top of this page. There's also a set of examples         
        <A HREF="http://svn.dsource.org/svn/projects/mango/trunk/example/">over here</A>. 
        Those will probably be sufficient to get an idea where the Mango Tree 
        might come in handy within your projects. 

        Discussion forums are hosted 
        <A HREF="http://www.dsource.org/forums/viewforum.php?f=5">here</A>, 
        and the general introduction starts 
        <A HREF="http://www.dsource.org/forums/viewtopic.php?t=148">here</A>. 

*******************************************************************************/




/******************************************************************************

        @file unittest.d

        Test jig for the mango.io package.


        This shows a number of different ways to manipulate buffers, 
        readers, writers, and how to perform IO on files and sockets.

        The basic premise behind this IO package is as follows:

        1) the central concept is that of a buffer. The buffer acts
           as a queue (line) where items are removed from the front
           and new items are added to the back. Buffers are modeled 
           by mango.io.model.IBuffer, and a concrete implementation 
           is provided via mango.io.Buffer.

        2) buffers can be written to directly, but a Reader and/or
           Writer are typically used to read & write formatted data.
           These readers & writers are bound to a specific buffer;
           often the same buffer. It's also perfectly legitimate to 
           bind multiple writers to the same buffer; they will all
           behave serially as one would expect. The same applies to
           multiple readers on the same buffer. Readers and writers
           support two styles of IO: put/get, and the C++ style << 
           and >> operators. All such operations can be chained.

        3) Any class can be made compatable with the reader/writer
           framework by implementing the IReadable and/or IWritable 
           interfaces. Each of these specify just a single method.

        4) Buffers may also be tokenized. This is handy when one is
           dealing with text input, and/or the content suits a more
           fluid format than most typical readers & writers support.
           Tokens are mapped directly onto buffer content, so there
           is only minor overhead in using them. Tokens can be read
           and written by reader/writers also, using a more relaxed
           set of rules than those applied to integral IO.

        5) buffers are sometimes memory-only, in which case there
           is nothing left to do when a reader (or tokenizer) hits
           end of buffer conditions. Other buffers are themselves 
           bound to a Conduit. When this is the case, a reader will 
           eventually cause the buffer to reload via its associated 
           conduit. Previous buffer content will thus be lost. The
           same concept is applied to writers, whereby they flush 
           the content of a full buffer to a bound conduit before 
           continuing. 

        6) conduits provide virtualized access to external content,
           and represent things like files or Internet connections.
           They are just a different kind of stream. Conduits are
           modelled by mango.io.model.IConduit, and implemented via
           classes FileConduit and SocketConduit. Additional kinds
           of conduit are easy to construct: one either subclasses
           mango.io.Conduit, or implements mango.io.model.IConduit. 
           The conduit reads and writes a buffer in large segments;
           typically the entire buffer.

        7) additional file-system support is provided through two
           classes: FileSystem supports operations such as getting 
           and setting the current directory, and FileProxy exposes
           facilities to manipulate both files and directories.

        8) console IO is implemented via Stdio, using FileConduit
           and appropriate readers and writers. Console IO can be
           redirected in the traditional (shell) manner.

        9) the package is intended to be portable between linux and
           Win32 platforms. 
        

        Kris Bell, March 26th 2004

*******************************************************************************/


        // test these modules
import  mango.io.Uri,
        mango.io.Stdio,
        mango.io.Token,
        mango.io.Reader,
        mango.io.Writer,
        mango.io.Buffer,
        mango.io.Socket,
        mango.io.Conduit,
        mango.io.FilePath,
        mango.io.FileStyle,
        mango.io.FileProxy,
        mango.io.Tokenizer,
        mango.io.TextReader,
        mango.io.TextWriter,
        mango.io.FileBucket,
        mango.io.FileConduit,
        mango.io.TextLayout,
        mango.io.TextFormat,
        mango.io.ColumnWriter,
        mango.io.DisplayWriter,
        mango.io.PickleReader,
        mango.io.PickleWriter,
        mango.io.SocketConduit,
        mango.io.SocketListener,
        mango.io.ArrayAllocator,
        mango.io.PickleRegistry,
        mango.io.DatagramSocket,
        mango.io.MulticastSocket;

        // need this for logging
import  mango.log.Logger,
        mango.log.Layout,
        mango.log.Manager,
        mango.log.DateLayout,
        mango.log.Configurator,
        mango.log.FileAppender,
        mango.log.StdioAppender,
        mango.log.ConsoleAppender,
        mango.log.RollingFileAppender;

        // need these for reading & writing classes
import  mango.io.model.IReader,
        mango.io.model.IWriter,
        mango.io.model.IPickle;

        // add a bunch of stuff for testing the http server
import  mango.http.server.HttpParams,
        mango.http.server.HttpCookies,
        mango.http.server.HttpHeaders,
        mango.http.server.HttpServer,
        mango.http.server.HttpRequest,
        mango.http.server.HttpResponse,       
        mango.http.server.HttpProvider;

import  mango.http.server.model.IProvider,
        mango.http.server.model.IProviderBridge;

import  mango.http.client.HttpClient;

import  mango.servlet.Servlet,
        mango.servlet.ServletConfig,
        mango.servlet.ServletContext,
        mango.servlet.ServletProvider;

import  mango.cache.PlainCache,
        mango.cache.HashMap,
        mango.cache.Payload,
        mango.cache.QueuedCache,
        mango.cache.VirtualCache;

import  mango.cluster.Message,
        mango.cluster.NetworkQueue,
        mango.cluster.NetworkCache,
        mango.cluster.CacheInvalidator,
        mango.cluster.CacheInvalidatee;
      
import  mango.cluster.qos.socket.Cluster,
        mango.cluster.qos.socket.ClusterServer,
        mango.cluster.qos.socket.ProtocolReader,
        mango.cluster.qos.socket.ProtocolWriter;

/*******************************************************************************

        global scope variables

*******************************************************************************/

private Logger logger;
        
/*******************************************************************************

        Simple test of in-memory buffer. Bind both a Reader and a Writer 
        onto the same buffer.
   
*******************************************************************************/

static void testBuffer()
{
        // create a small heap buffer
        Buffer buf = new Buffer (256);

        // map same buffer into both reader and writer
        Writer w = new Writer (buf);
        Reader r = new Reader (buf);

        int i = 10;
        int j = 20;
        double d = 3.14159;
        char[] c = "testing";
       
        // write integral types out
        w << c << i << j << d;

        // read them back again
        r >> c >> i >> j >> d;

        assert(i==10);
        assert(j==20);
        assert(d==3.14159);
        assert(c=="testing");

        // reset
        buf.clear();

        // same thing again, but using put() syntax instead (and swap i & j)
        w.put(c).put(i).put(j).put(d);
        r.get(c).get(j).get(i).get(d);

        assert(i==20);
        assert(j==10);
        assert(d==3.14159);
        assert(c=="testing");
}


/*******************************************************************************

        Append to a buffer directly without using a Writer. This can be
        useful for appending a set of strings (or other array type) to
        a contiguous space in a similar fashion to std.outbuffer.

*******************************************************************************/

static void testAppend()
{
        char[] foo = "to write some D code";

        // create a small heap buffer
        Buffer buf = new Buffer (256);

        // append some text directly to it
        buf.append("now is the time for all good men ").append(foo);

        // output the combined string
        Stdout.put (buf.toString).cr();
}


/*******************************************************************************

        write delimited text tokens to a buffer, then read them back out 
        to the console

*******************************************************************************/

static void testToken1()
{
        // create a small buffer on top of a static array
        static ubyte[128] storage;
        Buffer buf = new Buffer (storage);

        // write tokens with a comma between each
        IWriter w = new TextWriter (buf, ",");
        
        // bind a comma-tokenizer to a token instance
        BoundToken token = new BoundToken (Tokenizers.comma);

        // write some stuff to the buffer
        w.put("now is the time for all good men").put(3.14159);
        
        // read it back, and display as text
        while (token.next (buf))
               Stdout.put (token).cr();
}

/*******************************************************************************

        Use tokens directly with reader/writer also

*******************************************************************************/

static void testToken2()
{
        // create a small buffer on the heap
        Buffer buf = new Buffer (256);

        // write items with a comma between each
        IWriter w = new TextWriter (buf, ",");

        // write some stuff to the buffer
        w << "now is the time for all good men" << 3.14159;

        // bind a couple of tokens to a comma tokenizer
        ReaderToken text = new ReaderToken (Tokenizers.comma);
        ReaderToken number = new ReaderToken (Tokenizers.comma);
        
        // create any old reader since we only use it for handling tokens
        IReader r = new Reader (buf);

        // populate both tokens via reader 
        r >> text >> number;

        // print them to the console
        Stdout << text << ':' << number;
        Stdout.cr();
}

/*******************************************************************************

        use TextReader to convert text into integral types. This is akin 
        to what scanf does.

*******************************************************************************/

static void testToken3()
{
        // create a small buffer
        Buffer buf = new Buffer (256);

        // write tokens with a comma between each
        TextWriter tw = new TextWriter (buf, ",");

        // write some stuff to the buffer
        tw << 101 << 3.14159;

        // map a comma-tokenizing reader onto our buffer 
        TextReader tr = new TextReader (buf, Tokenizers.comma);

        real pi;
        int  integer;
        
        // convert both tokens via reader 
        tr >> integer >> pi;

        // print them to the console
        Stdout << integer << ',' << pi;
        Stdout.cr();
}


/*******************************************************************************

        Read line tokens via a TextReader

*******************************************************************************/

version (ArrayReaderSupport)
{
static void testToken4()
{
        // open a file for reading
        FileConduit fc = new FileConduit ("test.txt");

        // create a buffer for transferring content
        IBuffer buffer = fc.createBuffer ();

        // map a line-tokenizing reader onto our buffer 
        TextReader tr = new TextReader (buffer, Tokenizers.line);

        char[] ca;

        // convert both tokens via reader 
        tr.get (ca);

        // print them to the console
        Stdout.put (ca).cr();
}
}


/*******************************************************************************

        Use printf via a wrapper class

*******************************************************************************/

static void testFormat()
{       
        // create a Pattern instance
        TextFormat tf = new TextFormat (256);

        // write text to Stdout
        Stdout.put (tf.format ("there are %d arguments to every %.*s", 3, "contradiction")).cr();
}


/*******************************************************************************

        Layout text strings in an indexed fashion

*******************************************************************************/

static void testLayout()
{       
        // create a Layout instance
        TextLayout!(char) tl = new TextLayout!(char)(256);

        // write text to Stdout
        Stdout.put (tl.format ("%2 %1", "one", "two")).cr();
}


/*******************************************************************************

        Use interfaces to make any class reader and/or writer compatible 

*******************************************************************************/

static void testClassIO()
{
        // define a readable/writable class (via interfaces)
        class Foo : IReadable, IWritable
        {
                private int x = 11;
                private int y = 112;

                void write(IWriter writer)
                {
                        // use opShl() format here ...
                        writer << x << y;
                }

                void read(IReader reader)
                {
                        // use get() format this time ...
                        reader.get(x).get(y);
                }
        }

        // create a small heap buffer
        Buffer buf = new Buffer (256);

        // map reader/writer onto same buffer
        Writer w = new Writer(buf);
        Reader r = new Reader(buf);

        // construct a Foo
        Foo foo = new Foo();

        // write it
        w << foo;
        
        // read it back
        r >> foo;
}


/*******************************************************************************

        Use interfaces to make any class serializable. This will be used
        by mango.server to move class-instances around the network. Each 
        class is effectively a factory for itself.

        Note that this is not the same as Java; you can't ship code with
        the data ... one might perhaps use dll's or something to do that.

*******************************************************************************/

static void testClassSerialization()
{
        // define a serializable class (via interfaces)
        class Bar : IPickle, IPickleFactory
        {
                private int x = 11;
                private int y = 112;
        
                void write (IWriter writer)
                {
                        writer.put(x).put(y);
                }

                void read (IReader reader)
                {
                        reader.get(x).get(y);
                        assert (x == 11 && y == 112);
                }

                Object create (IReader reader)
                {
                        Bar bar = new Bar ();
                        bar.read (reader);
                        return bar;
                }

                char[] getGuid ()
                {
                        return "unittest.Bar";
                }
        }

        static Object create (IReader reader)
        {
                Bar bar = new Bar();
                bar.read (reader);
                return bar;
        }

        // setup for serialization
        Buffer buf = new Buffer (256);
        PickleWriter w = new PickleWriter (buf);
        PickleReader r = new PickleReader (buf);

        r.setAllocator (new BufferAllocator(r));

        // construct a Bar
        Bar bar = new Bar ();

        // tell registry about this object
        //PickleRegistry.enroll (&create, "unittest.Bar");
        PickleRegistry.enroll (bar);

        // serialize it
        w.freeze (bar);
        
        // create a new instance and populate. This just shows the basic
        // concept, not a fully operational implementation
        Object o = r.thaw ();
}


/*******************************************************************************

        open a file, and stream directly to console

*******************************************************************************/

static void testFile4()
{
        // open a file for reading
        FileConduit fc = new FileConduit ("test.txt");

        // stream directly to console
        Stdio.stdout.copy (fc);
}


/*******************************************************************************

       Create a file for random access. Write some stuff to it, rewind to
       file start and read back.

*******************************************************************************/

static void testRandomAccess()
{
        // open a file for reading
        auto FileConduit fc = new FileConduit ("random.bin", FileStyle.ReadWriteCreate);

        // construct (binary) reader & writer upon this conduit
        Reader r = new Reader(fc);
        Writer w = new Writer(fc);

        int x=10, y=20;

        // write some data
        w.put(x).put(y);

        // flush output since IO is buffered
        w.flush();

        // rewind to file start
        fc.seek (0);

        // read data back again, but swap destinations
        r.get(y).get(x);

        assert (y==10);
        assert (x==20);
}


/*******************************************************************************

       Create a file for random access. Write some stuff to it, rewind to
       file start and read back.

*******************************************************************************/

static void testConduitFilters()
{
        class MyFilter1 : ConduitSource
        {
                int reader (void[] dst)
                {
                        printf ("filtering input ...\n");
                        return next.reader (dst);
                }
        }

        class MyFilter2 : ConduitSink
        {
                int writer (void[] src)
                {
                        printf ("filtering output ...\n");
                        return next.writer (src);
                }
        }

        // open a file for reading
        auto FileConduit fc = new FileConduit ("random.bin", FileStyle.ReadWriteCreate);
        fc.attach (new MyFilter2);
        fc.attach (new MyFilter1);
        fc.attach (new MyFilter1);

        // construct (binary) reader & writer upon this conduit
        Reader r = new Reader(fc);
        Writer w = new Writer(fc);

        int x=1;

        // write some data
        w.put(x).flush();

        // rewind to file start
        fc.seek (0);

        // read data back again
        r.get(x);
}


/*******************************************************************************

        open a file, and explicitly buffer it to another conduit. This is
        effectively what Conduit.copyTo() does.

*******************************************************************************/

static void testFile5()
{
        // open a file for reading
        FileConduit fc = new FileConduit ("test.txt");

        // create a buffer for transferring content
        IBuffer buffer = fc.createBuffer ();

        // stream the data from file to stdout. This is probably as fast as 
        // one can practically make operations of this type. 
        while (fc.read (buffer) != fc.Eof)
               Stdio.stdout.write (buffer);

        // ensure the writer actually wrote everything
        Stdio.stdout.flush (buffer);
}


/*******************************************************************************

        Open a file and read it line-by-line using a line tokenizer. This 
        uses CompositeToken to hide some of the details. 

*******************************************************************************/

static void testFile1()
{
        // open a file for reading
        FileConduit fc = new FileConduit ("test.txt");

        // create a Token and bind it to both the file and a line-tokenizer
        CompositeToken line = new CompositeToken (Tokenizers.line, fc);

        // read file a line at a time. Method next() returns false when no more 
        // delimiters are found. Note there may be an unterminated line at eof
        while (line.get)
               Stdout.put(line).cr();
}


/*******************************************************************************

        Unwound version of above: we now explicitly pass a buffer instance 
        to the tokenizer, although the token itself is still bound.

*******************************************************************************/

static void testFile2()
{
        // open a file for reading
        FileConduit fc = new FileConduit ("test.txt");

        // create a buffer for reading the file
        IBuffer buffer = fc.createBuffer();

        // bind a line-tokenizer to our input token
        BoundToken line = new BoundToken (Tokenizers.line);

        // read file a line at a time. Method next() returns false when no more 
        // delimiters are found. Note there may be an unterminated line at eof
        while (line.next(buffer) || line.getLength())
               Stdout.put(line).cr();
}


/*******************************************************************************

        Further unwinding of above: we invoke the tokenizer directly, 
        explicitly passing both the buffer and the token

*******************************************************************************/

static void testFile3()
{
        // open a file for reading
        FileConduit fc = new FileConduit ("test.txt");

        // create a buffer for reading the file
        IBuffer buffer = fc.createBuffer();

        // create a token for receiving the line
        Token token = new Token ();

        // read file a line at a time. Method next() returns false when no more 
        // delimiters are found. Note there may be an unterminated line at eof
        while (Tokenizers.line.next(buffer, token) || token.getLength())
               Stdout.put(token).cr();
}


/*******************************************************************************

        Recursively scan files and directories, adding filtered files to
        an output structure as we go. Thanks to Chris S for this.
        
*******************************************************************************/

void testScanFiles()
{       
        /***********************************************************************

        ***********************************************************************/

        struct Dependencies 
        {
                FileProxy[]     mods;
                FilePath[]      pkgs;
        }

        /***********************************************************************

        ***********************************************************************/

        void scanFiles (inout Dependencies deps, FilePath base) 
        {
                /***************************************************************

                        delegate for filtering sub-directories and 
                        D source files

                ***************************************************************/

                bit filter (FilePath fp) 
                {
                        char[]  sbuf = fp.getExtension();

                        if (fp.getName[0] != '.')                       
                            if (sbuf.length == 0 || sbuf == "d")
                                return true;

                        return false;
                }

                /***************************************************************

                ***************************************************************/

                FileProxy  proxy = new FileProxy (base);
                FilePath[] paths = proxy.toList  (&filter);

                // add packages only if there's something in them
                if (paths.length)
                    deps.pkgs ~= base;

                foreach (FilePath x; paths) 
                        {
                        // combine base path with listed file
                        FilePath spliced = new FilePath (x.splice (base));

                        // recurse if this is a directory ...
                        proxy = new FileProxy (spliced);
                        if (proxy.isDirectory) 
                            scanFiles (deps, spliced);
                        else 
                           deps.mods ~= proxy;
                        }
        }


        /***********************************************************************

        ***********************************************************************/

        Dependencies deps;
        scanFiles (deps, new FilePath(r"\dmd\src\mango"));

        Stdout.put ("-----------------").cr();
        foreach (FileProxy mod; deps.mods)
                 Stdout.put (mod.getPath()).cr();

        Stdout.put ("-----------------").cr();
        foreach (FilePath pkg; deps.pkgs)
                 Stdout.put (pkg).cr();
}


/*******************************************************************************

        Open an internet connection, and copy home page to console (via
        Chris Miller's socket.d). We could use a line-by-line approach in
        the same manner as above since both Files & Sockets are conduits. 
        All conduits are interchangable with others.

*******************************************************************************/

static void testSocket()
{
        // create a socket and connect it to Walter's site
        SocketConduit sc = new SocketConduit();
        sc.connect (new InternetAddress("www.digitalmars.com", 80));

        // construct a (flushing) writer and bind it to the socket
        IWriter w = new DisplayWriter (new FlushBuffer(1024, sc));

        // send HTTP request
        w << "GET /d/intro.html HTTP/1.1\r\n"
             "Host: www.digitalmars.com\r\n"
             "\r\n";

        // set read-timeout to 0.5 seconds (avoids stalling)
        sc.setTimeout (System.Interval.Millisec * 500);

        // stream home-page directly to console conduit
        Stdio.stdout.copy (sc);
}


/*******************************************************************************

        Use a ColumnWriter to write formatted data onto the console. Note
        that we map the ColumnWriter directly into the Stdout buffer.

*******************************************************************************/

static void testColumn()
{
        static int[] columns = [0, 4, 14, 24];

        // map a ColumnWriter directly onto Stdout
        ColumnWriter cw = new ColumnWriter (Stdout.getBuffer(), new ColumnList(columns));

        // set 2 digits of precision
        cw.setPrecision (2);

        // put() style
        cw.put(1);
        cw.put(20.34);
        cw.put("test");
        cw.cr();
        cw.put("next");
}

/*******************************************************************************

        Get a line of input from the console, and extract an integer. Input
        from the console always waits until CR is pressed (at the OS level).

*******************************************************************************/

static void testStdin()
{
        int x;
        
        // display a prompt
        Stdout << "Enter a number: ";

        // wait for <cr>, then convert input to integer.
        Stdin >> x;
}


import std.regexp;
        
/*******************************************************************************

        Use RegExp to scan words via a buffer. 

*******************************************************************************/

static void wordRegex (IBuffer buffer)
{
        // create a new regex tokenizer for words
        RegexTokenizer rt = new RegexTokenizer (new RegExp(r"\w+", null));

        // create a token to hold tokenized output, and bind it to the buffer
        CompositeToken word = new CompositeToken (rt, buffer);

        // read buffer a word at a time. Method next() returns false when no 
        // more tokens are found
        while (word.next)
               Stdout.put(word).cr();
}



/*******************************************************************************

        Use RegExp to scan file content via a buffer. Buffered scanning  
        avoids loading the entire file into memory at one time.

*******************************************************************************/

static void testFileRegex()
{
        // open a file for reading
        FileConduit fc = new FileConduit ("test.txt");
        
        // extract words from file
        wordRegex (fc.createBuffer);
}


/*******************************************************************************

        open an internet connection and scan home-page for words. 

*******************************************************************************/

static void testSocketRegex()
{
        // create a socket and connect it to Walter's site
        SocketConduit sc = new SocketConduit();
        sc.connect (new InternetAddress("www.digitalmars.com", 80));

        // construct a (flushing) writer and bind it to the socket
        IWriter w = new DisplayWriter (new FlushBuffer(256, sc));

        // send HTTP request
        w << "GET /d/intro.html HTTP/1.1\r\n"
             "Host: www.digitalmars.com\r\n"
             "\r\n";

        // set read-timeout to 0.5 seconds (avoids stalling)
        sc.setTimeout (System.Interval.Millisec * 500);

        // scan words out of home page
        wordRegex (sc.createBuffer);
}


/*******************************************************************************

        Read lines from a file, and scan each line for words. This shows
        how to map the content of one buffer into another without copying
        any data. Unlikely to be a common activity.
        
*******************************************************************************/

static void testLineRegex()
{
        // open a file for reading
        FileConduit fc = new FileConduit ("test.txt");
        
        // create a token for lines, and map it onto the file buffer.
        CompositeToken line = new CompositeToken (Tokenizers.line, fc);

        // create a seperate buffer for scanning within each line. We will 
        // set the content for this buffer as we read each line. Note that
        // if we used the file buffer directly, then we'd continue reading
        // words right past the end of the line. This is because the file
        // buffer would just reload itself instead of halting at the line
        // termimator.
        Buffer rescan = new Buffer;

        // scan all lines in the file, including the final one
        while (line.get)
              {
              // map our rescan buffer onto the line content 
              rescan.setValidContent (line.toString);

              // scan just this line for words
              wordRegex (rescan);
              }
        
}

/*******************************************************************************

        Display the size of a file 
        
*******************************************************************************/

static void testFileSize()
{
        FileProxy fp = new FileProxy ("test.txt");

        Stdout.put (fp.getPath)
              .put(" is ")
              .put(fp.getSize)
              .put(" bytes long")
              .cr();
}                


/*******************************************************************************

        list a set of files
        
*******************************************************************************/

static void testFileList()
{
        FileProxy proxy = new FileProxy (r".");

        foreach (FilePath path; proxy.toList())
                 Stdout.put(path).cr();
}                

/*******************************************************************************

        list a set of files
        
*******************************************************************************/

static void testFilePath()
{
        FilePath path = new FilePath (r"c:\1\2\3\4\abc.foo.bar");

        Stdout << path << Writer.newline;

        Uri uri = path.toUri();
        Stdout << uri.getScheme() << Writer.newline
               << uri.getHost() << Writer.newline
               << uri.getPath() << Writer.newline
               << uri.getPort() << Writer.newline
               << uri << Writer.newline;

        uri = new Uri (uri.toString());
        Stdout << uri.getScheme() << Writer.newline
               << uri.getHost() << Writer.newline
               << uri.getPath() << Writer.newline
               << uri.getPort() << Writer.newline
               << uri << Writer.newline;

        path = new FilePath (uri);

        Stdout << path << Writer.newline;
}                


/*******************************************************************************

        Construct a memory-mapped buffer upon a 200MB file. Note that we 
        should close both the conduit and buffer when finished. 

        Note also that Win32 takes a while to close the initial file if
        it has just been created or resized. Subsequently opening the
        file as ReadWriteExising performs admirably. I imagine it's to
        do with spooling a gazillion dirty pages to the drive?

*******************************************************************************/

void testMappedFile()
{
        FileConduit conduit = new FileConduit ("test.bin", FileStyle.ReadWriteCreate);
        conduit.seek (200_000_000);
        conduit.truncate ();

        MappedBuffer buffer = conduit.createMappedBuffer;
        buffer.setPosition (100_000_000);
        buffer.append ("test");
        buffer.close ();

        conduit.close ();
}


/*******************************************************************************

        Use Composite IO to read and write a data-set. See CompositeReader
        and CompositeWriter for more on this.
         
*******************************************************************************/

static void testCompositeIO()
{
        // define a serializable class (via interfaces)
        class Wumpus : IReadable, IWritable
        {
                private int     x = 11,
                                y = 112,
                                z = 1024;

                void write(IWriter writer)
                {
                        writer.put(x).put(y).put(z);
                }

                void read(IReader reader)
                {
                        reader.get(x).get(y).get(z);
                }

        }

        // construct a Wumpus
        Wumpus wumpus = new Wumpus();

        // open a file for IO
        FileConduit fc = new FileConduit ("random.bin", FileStyle.ReadWriteCreate);

        // construct composite reader & writer upon the file, with binary IO
        IWriter w = new Writer(fc);
        IReader r = new Reader(fc);

        // write the Wumpus (and flush it)
        w.put(wumpus).flush();

        // rewind to file start
        fc.seek (0);

        // read Wumpus back again
        r.get (wumpus);

        fc.close();
}



/*******************************************************************************

        One can perform direct IO on a conduit, simply by mapping a Buffer
        onto a local array. No additional buffering takes place but readers, 
        writers, and tokenizers can't tell the difference: they may all be 
        used without issue.

        This example uses the same direct-buffer for both reading & writing.
        However, seperate buffers may be used as appropriate.

*******************************************************************************/

static void testDirectIO()
{
        // open a file for IO
        auto FileConduit fc = new FileConduit ("random.bin", FileStyle.ReadWriteCreate);

        // direct IO area (on the stack)
        ubyte[1024] content;

        // map a buffer onto local content. State all of it is valid content.
        // There are other IBuffer methods, such as setValidContent(), which
        // stipulate how much data is really in the buffer.
        IBuffer direct = new Buffer(content, content.length);

        // map our buffer onto the file
        direct.setConduit(fc);

        // mess with the content[] ...
        // ...
        // ...
        // ...

        // write directly from the local buffer (might use fc.flush() instead)  
        fc.write (direct);

        // rewind to file start
        fc.seek (0);

        // make space for reading 
        direct.clear();

        // read it all back in again
        int count = fc.read (direct);
        delete fc;

}


/*******************************************************************************

*******************************************************************************/

static void testUri()
{
        MutableUri uri;
        
        uri = new MutableUri ("ftp1://Me:My@a:21");
        Stdout << uri << Writer.newline;

        uri.parse ("http://a:8080/b/c/d;p?q=bar of soap#fragmented");
        Stdout << uri << Writer.newline;

        Stdin.wait();

        for (int i=10_000_000; i > 0; --i)
             uri.parse ("http://a:8080/b/c/d;p?q=bar of soap#fragmented");
}


/******************************************************************************
************************ SocketServer testing *********************************
*******************************************************************************/



/*******************************************************************************
        
        Test the http server by sending it a few thousand requests.  Note
        that we don't use a reader or a writer here, though we could just
        as easily have done so.

        Also note that (Gentoo) linux has a hard time scavenging sockets
        fast enough to complete more than around 6000 iterations. Without
        the appropriate guru incantations to make ulimit work as desired,
        we just limit the iterations to a more agreeable level. Testing 
        under Win32 seems to be fine at the 100,000 iteration mark.

*******************************************************************************/

void testClient ()
{
        Buffer buf = new Buffer(1024);
        InternetAddress addr = new InternetAddress("127.0.0.1", 80);

        for (int i=50000; --i >= 0;)
            {
            // connect to the server
            SocketConduit socket = new SocketConduit ();
            socket.connect (addr);

            // when we close this socket, recycle it immediately!
            socket.setLingerPeriod (0);

            // reset buffer for another go around
            buf.clear ();
                
            // would need this if we were using Reader/Writer IO
            // buf.setConduit(socket);

            // send a request. This will be serviced by our IProvider
            buf.append ("GET / HTTP/1.1\n\n");
/+
            buf.append("POST /RPC2 HTTP/1.1\n"
            "Content-Length: 157\n"
            "Content-Type: text/xml\n"
            "Cache-Control: no-cache\n"
            "Pragma: no-cache\n"
            "User-Agent: Java/1.4.2_05\n"
            "Host: localhost:8181\n"
            "Accept: text/html, image/gif, image/jpeg, *; q=.2, */*; q=.2\n"
            "Connection: keep-alive\n\n"
            "<?xml version=\"1.0\" encoding=\"ISO-8859-1\"?><methodCall>"
            "<methodName>nnCore.getNode</methodName><params><param>"
            "<value>1@1</value></param></params></methodCall>");
+/
            socket.write (buf);

            // wait for something to be returned (you can set a timeout)
            socket.read (buf);

            // dump response to the console
            //Stdio.stdout.write (buf);

            // attempt a graceful shutdown
            socket.shutdown ();

            // close and destroy socket (the latter will do both)
            socket.close ();
            delete socket;
            }
}


/*******************************************************************************

        Create an http server with the given IProvider

*******************************************************************************/

void testServer (IProvider provider)
{       
        // bind to port 80 on a local address
        InternetAddress addr = new InternetAddress (80);

        // create a (1 thread) server using the IProvider to service requests
        HttpServer server = new HttpServer (provider, addr, 1);

        // start listening for requests (but this thread does not listen)
        server.start ();
        
        // pause for some console input
        Stdout << "hit return to start the client test ...";
        Stdin.wait ();

        // send a few thousand requests
        long time = System.getMillisecs;
        testClient ();
        Stdout.put (System.getMillisecs - time).cr();
}


/*******************************************************************************

        Declare a class for handling http requests. This one just simply
        returns a status page

*******************************************************************************/

void testHttpServer ()
{       
        class Provider : HttpProvider
        {
                void service (HttpRequest request, HttpResponse response)
                {
                        response.sendError (HttpResponses.OK);
                }
        }

        // create a new http-provider, and attach it to a server
        testServer (new Provider());
}


/*******************************************************************************

        Test the servlet wrapper. Point your Browser at http://127.0.0.1
        and optionally add a path, query params, etc ...

*******************************************************************************/

void testServletEngine ()
{       
        /***********************************************************************

                Return a file. Deliberately supports GET requests only. 

        ***********************************************************************/

        class FileServlet : MethodServlet
        {
                void doGet (IServletRequest request, IServletResponse response)
                {   
                        response.copyFile (request.getContext(), request.getPathInfo());
                }
        }


        /***********************************************************************

                A whole bunch of diagnostic stuff, plus cookie return and
                rudimentary form processing.

        ***********************************************************************/

        class MyServlet : Servlet
        {
                void service (IServletRequest request, IServletResponse response)
                {   
                        Uri uri = request.getUri();

                        // log everything to the console
                        Stdout.put ("------------------------")
                              .cr  ()
                              .put ("Uri: ")
                              .put (uri)
                              .cr  ()
                              .put ("------------------------")
                              .cr  ()
                              .put ("Headers:")
                              .cr  ()
                              .put (request.getHeaders())
                              .put ("------------------------")
                              .cr  ()
                              .put ("Cookies:")
                              .cr  ()
                              .put (request.getCookies())
                              .put ("------------------------")
                              .cr  ()
                              .put ("Parameters:")
                              .cr  ()
                              .put (request.getParameters())
                              .put ("------------------------")
                              .cr  ();

                        // display the Servlet environment
                        Stdout.put("encoding: ")
                              .put(request.getCharacterEncoding())
                              .cr ()
                              .put("content length: ").put(request.getContentLength())
                              .cr ()
                              .put("content type: ").put(request.getContentType())
                              .cr ()                   
                              .put("protocol: ").put(request.getProtocol())
                              .cr ()                   
                              .put("scheme: ").put(uri.getScheme())
                              .cr ()                   
                              .put("method: ").put(request.getMethod())
                              .cr ()                   
                              .put("host name: ").put(request.getServerName())
                              .cr ()                   
                              .put("host port: ").put(request.getServerPort())
                              .cr ()                   
                              //.put("remote address: ").put(request.getRemoteAddress()).cr()                   
                              //.put("remote host: ").put(request.getRemoteHost()).cr()                   
                              .put("path info: ").put(request.getPathInfo())
                              .cr ()                   
                              .put("query: ").put(uri.getQuery())
                              .cr ()                   
                              .put("path: ").put(uri.getPath())
                              .cr ()                   
                              .put("context path: ").put(request.getContextPath())
                              .cr ()                   
                              .cr ()
                              .cr ();  

                        int i = request.getContentLength();

                        if (i > 0)
                           {
                           byte[] content = new byte[i];

                           request.getReader.getPostData (content);
                           printf ("Content:\n>>%.*s<<\n\n\n", content);
                           }

                        // send a cookie to the user-agent ...
                        response.getCookies().add (new Cookie ("fu", "'bar of soap'"));

                        // send a form to the user-agent ...
                        IWriter writer = response.getWriter();
                        writer.put ("<HTML><HEAD><TITLE>Form</TITLE></HEAD>")
                              .put ("<BODY><FORM method=POST action=\"/context/test\">")
                              .put ("Enter a POST parameter: ")
                              .put ("<INPUT type=text name=\"name\"><P>")
                              .put ("<INPUT type=hidden name=\"hidden\" value=\"test\"><P>")
                              .put ("<INPUT type=submit>")
                              .put ("<?FORM></BODY></HTML>")
                              .cr  ();

                }
        }
          

        // construct a servlet-provider
        ServletProvider sp = new ServletProvider();

        // map all html requests to our file servlet
        IRegisteredServlet files = sp.addServlet (new FileServlet(), "files");
        sp.addMapping ("*.html", files);
        sp.addMapping ("*.htm", files);
        
        // map all other request to our test servlet
        IRegisteredServlet test = sp.addServlet (new MyServlet(), "test");
        sp.addMapping ("/", test);

        // fire up a server
        testServer (sp);
}


/*******************************************************************************

        Shows a variety of ways in which to access data returned from
        a client-side socket connection.

*******************************************************************************/

void testHttpClient ()
{  
        /***********************************************************************

                Bulk copy from input to Stdout

        ***********************************************************************/

        void readBulk (IConduit conduit)
        {
                // stream directly to console
                Stdio.stdout.copy (conduit);
        }

        /***********************************************************************

                Unwound version of the above, where we explicitly move
                data between endpoints using a buffer

        ***********************************************************************/

        void readBulkExplicit (IConduit conduit)
        {
                // create a buffer for transferring content
                IBuffer buffer = conduit.createBuffer ();

                // stream the data from file to stdout. This is probably as fast as 
                // one can practically make operations of this type. 
                while (conduit.read (buffer) != conduit.Eof)
                       Stdio.stdout.write (buffer);

                // ensure the writer actually wrote everything
                assert (Stdio.stdout.flush (buffer));
        }

        /***********************************************************************

                Read input a line at a time from the input. You can access
                the line content via method Token.toString()

        ***********************************************************************/

        void readLines (IConduit conduit)
        {
                // create a Token and bind it to both the conduit and a line-tokenizer
                CompositeToken line = new CompositeToken (Tokenizers.line, conduit);

                // read data a line at a time. Method next() returns false when no more 
                // delimiters are found. Note there may be an unterminated line at eof
                while (line.get)
                       Stdout.put(line).cr();
        }

        /***********************************************************************

                Read input a char at a time. This will read in big chunks
                of data from the input, but spit them out one character at
                a time. While this is a lot faster then reading one char
                at a time from the OS (the typical approach), this style
                of reading is not recommended. Instead, try the 'chunked'
                approach (below).

        ***********************************************************************/

        void readChars (IConduit conduit)
        {
                // create a reader on the conduit
                Reader reader = new Reader (conduit);

                // read a single character at a time, until we run out of them.
                // An exception will eventually be thrown because we're using
                // the 'formatted' data approach
                try {
                    while (true)
                          {
                          char c;
                          reader.get (c);
                          Stdout.put (c);
                          }
                    } catch (IOException x){}
        }

        /***********************************************************************

                Read big chunks of data, and process them as such. Each read
                operation will eat as much as there is from the input, up to
                the size limit of the buffer.

        ***********************************************************************/

        void readChunks (IConduit conduit)
        {
                IBuffer buffer = new Buffer(1024);        
        
                while (conduit.read (buffer) != conduit.Eof)
                      {
                      Stdout.put (buffer.toString);
                      buffer.clear ();
                      }
        }

        /***********************************************************************
        
                unwound version of the above, where we setup buffer space
                on the stack (instead of the heap). We need to explicitly
                ask the buffer how much data is actually available. One
                could use the return value from read() as the number of
                bytes available instead (in this particular case), or 
                use buffer.toString() as in the above case

        ***********************************************************************/

        void readChunksExplicit (IConduit conduit)
        {
                char[1024] content;

                IBuffer buffer = new Buffer (content);        

                while (conduit.read (buffer) != conduit.Eof)
                      {
                      Stdout.put (content[0..buffer.readable()]);
                      buffer.clear ();
                      }
        }


        /***********************************************************************

                Open a web-site connection, and process the returned page

        ***********************************************************************/

        // create a socket and connect it to Walter's site
        SocketConduit sc = new SocketConduit();
        sc.connect (new InternetAddress("www.digitalmars.com", 80));

        // construct a (flushing) writer and bind it to the socket
        IWriter w = new DisplayWriter (new FlushBuffer (256, sc));
        
        // send HTTP request
        w.put ("GET /d/intro.html HTTP/1.1")
         .cr  ()
         .put ("Host: www.digitalmars.com")
         .cr  ()
         .cr  ();

        // set read-timeout to 1 second (avoids stalling for ever!)
        sc.setTimeout (System.Interval.Second);
        
        //readBulk (sc);
        //readBulkExplicit(sc);
        //readLines(sc);
        //readChars(sc);
        //readChunks(sc);
        readChunksExplicit(sc);
}
     

/*******************************************************************************

        Extract client-side headers from an HTTP reply
        
*******************************************************************************/

void testHttpClient2 ()
{  
        // create a socket and connect it to Walter's site
        SocketConduit sc = new SocketConduit;
        sc.connect (new InternetAddress("www.digitalmars.com", 80));

        // construct a (flushing) writer and bind it to the socket
        IWriter w = new DisplayWriter (new FlushBuffer (256, sc));
        
        // send HTTP request
        w.put ("GET /d/intro.html HTTP/1.1")
         .cr  ()
         .put ("Host: www.digitalmars.com")
         .cr  ()
         .cr  ();

        // set read-timeout to 1 second (avoids stalling for ever!)
        sc.setTimeout (System.Interval.Second);

        // create an input buffer
        IBuffer buffer = sc.createBuffer;
        
        // extract all headers 
        HttpHeaders headers = new HttpHeaders;
        headers.parse (buffer);

        // display parsed headers -- these should match HttpHeader.HttpHeaders
        foreach (HeaderElement header; headers)
                 Stdout.put (header.name.value)
                       .put (header.value)
                       .cr  ();

        // display remaining content
        while (sc.read (buffer) != sc.Eof)
              {
              Stdout.put (buffer.toString);
              buffer.clear ();
              }
}

/*******************************************************************************

        Poke an HTTP server, using HttpClient.

*******************************************************************************/

void testHttpClient3()
{
        // callback for client reader
        void sink (char[] content)
        {
                Stdout.put (content);
        }

        // create client for a GET request
        auto HttpClient client = new HttpClient (HttpClient.Get, "http://www.digitalmars.com/d/intro.html");

        // setup a Host header
        client.getRequestHeaders.add (HttpHeader.Host, client.getUri.getHost);

        // make request
        client.open ();

        // check return status for validity
        if (client.isResponseOK)
           {
           // extract content length (be aware of -1 return, for no header)
           int length = client.getResponseHeaders.getInt (HttpHeader.ContentLength);
           if (length < 0)
               length = int.max;
        
           // display all returned headers
           Stdout.put (client.getResponseHeaders);
        
           // display remaining content
//           client.read (&sink, length);
           client.read (delegate (char[] c){Stdout.put(c);}, length);
           }
        else
           Stderr.put (client.getResponse);
}




/******************************************************************************
**************************** Cache testing ************************************
*******************************************************************************/



/*******************************************************************************


*******************************************************************************/

void testPlainCache()
{
        PlainCache cache = new PlainCache (2);

        cache.put ("a", new Payload);
        cache.put ("b", new Payload);
        cache.put ("c", new Payload);
        cache.put ("c", new Payload);

        cache.extract ("b");
        assert (cache.get("a"));
        assert (cache.get("c"));
        assert (cache.get("b") is null);
}

/*******************************************************************************

        QueuedCache maintains a linked list of items, in addition to a
        hashmap. Any reference to an item places it at the list head, 
        resulting in the LRU items always being left at the tail. When
        the list becomes full, tail items are dropped and their places
        reused.

        This is great for keeping commonly accessed items around, while
        limiting the amount of memory used.

        Typically, the queue size would be set in the hundreds (perhaps
        thousands) but we just use 2 for testing purposes. 

*******************************************************************************/

void testQueuedCache()
{
        QueuedCache cache = new QueuedCache (2);

        cache.put ("a", new Payload);
        cache.put ("b", new Payload);
        cache.put ("c", new Payload);
        cache.put ("c", new Payload);

        assert (cache.get("a") is null);
        assert (cache.get("c"));
}

/*******************************************************************************

        Stuff some data into a FileBucket, and pull it out again.

*******************************************************************************/

void testFileBucket()
{
        char[] text = "this is a test";

        FileBucket bucket = new FileBucket (new FilePath("bucket.bin"), FileBucket.HalfK, 0);

        // insert some data, and retrieve it again
        bucket.put ("a", text);
        char[] b = cast(char[]) bucket.get ("a");

        assert (b == text);
        bucket.close();
}


/*******************************************************************************

        Use a combination of QueuedCache and FileBucket to spill LRU
        cache entries onto the hard-drive, and then recover them intact.

        This demonstrates what one can do with the IPickle notion.

*******************************************************************************/

void testVirtualCache()
{
        FileBucket bucket = new FileBucket (new FilePath("bucket.bin"), FileBucket.HalfK);
        VirtualCache cache = new VirtualCache (bucket, 2);
        PickleRegistry.enroll (new Payload);
        
        cache.put ("a", new Payload);
        cache.put ("b", new Payload);
        cache.put ("c", new Payload);

        assert (cache.get("a"));
        assert (cache.get("c"));
        assert (cache.get("b"));
        bucket.close();
}

/*******************************************************************************


*******************************************************************************/

void testProtocol()
{
        class Protocol : Payload
        {
                char[]  x;
                int     y = 10;
                bool    z = true;
        
                this ()
                {
                        x = "fred";
                }

                static this()
                {
                        PickleRegistry.enroll (new Protocol);
                }

                void write (IWriter w)
                {
                        super.write (w);
                        w.put (x)
                         .put (y)
                         .put (z);
                }

                void read (IReader r)
                {
                        super.read (r);
                        r.get (x)
                         .get (y)
                         .get (z);

                        assert (x == "fred");
                        assert (y == 10);
                        assert (z == true);
                }

                Payload create ()
                {
                        return new Protocol;
                }

                char[] getGuid ()
                {
                        return this.classinfo.name;
                }
        }

        Protocol p = new Protocol;

        Buffer b = new Buffer (1024);
        ProtocolWriter w = new ProtocolWriter (b);
        ProtocolReader r = new ProtocolReader (b);

        w.put (w.Command.Exception, "mychannel", p);

        ubyte   cmd;
        char[]  channel,
                element;

        p = cast(Protocol) r.getPayload (channel, element, cmd);
}


/*******************************************************************************


*******************************************************************************/

void testHashMap()
{
        class AAPut
        {
                Object[char[]] aa;

                synchronized void put (char[] k, Object v)
                {
                        aa[k] = v;
                }

                synchronized Object get (char[] k)
                {
                        if (k in aa)
                            return aa[k];
                        return null;
                }

                synchronized void del (char[] k)
                {
                        delete aa[k];
                }
        }


        static const char[] key = "dkfsdflksdjf s;dklf kdfjf;lskdfjs d;flskdfjs dfl";

        AAPut aa = new AAPut;
        HashMap map = new HashMap (1024);

        map.put (key, Client.EmptyMessage);
        aa.put (key, Client.EmptyMessage);

        map.put ("asdfasdfasdfasdfasdfasdfasdfasdfasdfasdf", Client.EmptyMessage);
        aa.put ("asdfasdfasdfasdfasdfasdfasdfasdfasdfasdf", Client.EmptyMessage);

        map.put ("asdfasdfasdfasdfasdfasdfasdfasdfasdfasdf 34534 5345", Client.EmptyMessage);
        aa.put ("asdfasdfasdfasdfasdfasdfasdfasdfasdfasdf 3453453535", Client.EmptyMessage);

        logger.info ("A");
        for (int i=5_000_000; --i;)
            {
            map.put (key, Client.EmptyMessage);
            map.remove (key);
            //map.get (key);
            }

        logger.info ("B");
        for (int i=5_000_000; --i;)
            {
            aa.put (key, Client.EmptyMessage);
            aa.del (key);
            //aa.get (key);
            }
/*

        map.put ("x", Client.EmptyMessage);
        map.put ("y", Client.EmptyMessage);
        assert (map.get ("a"));
        assert (! map.get ("b"));
        assert (map.remove ("a"));
        assert (! map.get ("a"));
*/
}


/*******************************************************************************


*******************************************************************************/

void testRollingFileLog()
{
        Logger l = Logger.getLogger ("test");
        l.addAppender (new RollingFileAppender (r"e:\test.log", 3, 1024, new DateLayout));

        for (int i=100; --i;)
             l.error ("weqwer er wer werqwer wer wer wer wr wqrwe qrw rwe rwqerw");
}




/******************************************************************************
**************************** Cluster testing **********************************
*******************************************************************************/



/*******************************************************************************

        Join a socket to a multicast group, and send the group a packet.

*******************************************************************************/

void testMulticast()
{
        class MyListener : SocketListener
        {
                this (ISocketReader reader, IBuffer buffer)
                {
                        super (reader, buffer);
                }

                void notify (IBuffer buffer)
                {
                        printf ("listener received %d bytes\n", buffer.readable());
                }

                void exception (char[] msg)
                {       
                        printf ("listener error: %.*s\n", msg);
                }
        }

        //auto MulticastSocket ms = new MulticastSocket ();
        MulticastSocket ms = new MulticastSocket ();
        InternetAddress ia = new InternetAddress ("226.1.1.1", 4444);

        ms.join (ia);
        MyListener s = new MyListener (ms, new Buffer (2000));
        s.start ();
        
        Stdin.wait ();
        //s.pause ();
        //s.resume ();

        byte[1000] xx;
        Buffer b = new Buffer (xx, xx.length);

        ms.write (b, ia);
        Stdin.wait ();
        //s.cancel ();
}

/*******************************************************************************

*******************************************************************************/

void testClusterCache()
{
        //ClusterServer cs = new ClusterServer (new InternetAddress(4567), 1);
        //cs.start ();

        //auto FileConduit config = new FileConduit ("cluster.properties");
        //Cluster cluster = new Cluster (logger, config);
        Cluster cluster = new Cluster (logger, 4567);

        //NetworkCache cc = new NetworkCombo (cluster, "myChannel", new PlainCache);
        NetworkCache cc = new NetworkCache (cluster, "myChannel");

        InvalidatorPayload p = new InvalidatorPayload;

            cc.put ("key", p);
//            cc.invalidate ("key");
        for (int i=100_000; i; --i)
            {
//            p.setText ("fubar");
//            cc.put ("key", p);
//            cc.invalidate ("key");
            IPayload p = cc.get ("key");
//            if ((i & 0x3f) == 0)
//                 System.sleep (System.Interval.Second);
/*
            p = cast(InvalidatorPayload) cc.get ("key");
            assert (p);
            assert (p.getText == "fubar");

            cc.invalidate ("key");
            p = cast(InvalidatorPayload) cc.get ("key");
*/
            }
        //assert (p is null);

//        cc.invalidate ("key");

        //Stdin.wait ();
}


/*******************************************************************************

*******************************************************************************/

void testClusterQueue()
{       
        class Listen : IEventListener
        {
                void notify (IEvent event, IPayload payload)
                {
                        printf ("recieved queue entry from channel '%.*s'\n", event.getChannel.getName);       
                }
        }

        //ClusterServer cs = new ClusterServer (new InternetAddress(4567), 1);
        //cs.start ();

        Cluster cluster = new Cluster (logger, 4567);
        NetworkQueue queue = new NetworkQueue (cluster, "queue.channel");

        queue.createConsumer (new Listen);
        queue.put (new InvalidatorPayload);
        queue.put (new InvalidatorPayload);
        
        Stdin.wait ();
}


/*******************************************************************************

*******************************************************************************/

void testInvalidatee()
{
        ICluster c = new Cluster (logger);

        CacheInvalidatee dst = new CacheInvalidatee (c, "abc", new PlainCache);
        CacheInvalidator src = new CacheInvalidator (c, "abc");

        src.invalidate ("key1");

        Stdin.wait ();
}


/*******************************************************************************


*******************************************************************************/

void testClusterServer()
{
        ClusterServer cs = new ClusterServer (new InternetAddress(81), 1);
        SocketConduit sc = new SocketConduit ();
        cs.start();

        sc.connect (new InternetAddress ("127.0.0.1", 81));
        Stdin.wait ();
}


/*******************************************************************************


*******************************************************************************/


void testClusterMessage()
{
        class MessageListener : IEventListener
        {
                void notify (IEvent event, IPayload payload)
                {
                        IMessage message = cast(IMessage) payload;

                        printf ("Replying to message\n");
                        event.reply (message.getReply, new InvalidatorPayload);
                }
        }

        class ReplyListener : IEventListener
        {
                void notify (IEvent event, IPayload payload)
                {
                        printf ("Received reply\n");
                }
        }

        //ClusterServer cs = new ClusterServer (new InternetAddress(4567), 1);
        //cs.start ();

        // open the cluster
        Cluster cluster = new Cluster (logger, 4567);

        // open up a message channel
        NetworkMessage msg = new NetworkMessage (cluster, "message.channel", new ReplyListener);

        // toss a message out to the cluster
        msg.put (new NullMessage);

        // setup a listener to receive message (could do this earlier)
        msg.createConsumer (new MessageListener);
        
        // wait for traffic to pass by
        Stdin.wait ();
}


/*******************************************************************************


*******************************************************************************/

void testCacheLoader()
{
        class CacheLoader : ICacheLoader
        {
                bool test (IPayload p)
                {
                        return false;
                }            
                
                IPayload load (char[] key, long time)
                {
                        printf ("loading local cache instance\n");
                        return new NullMessage;
                }    
        }

        IMutableCache mc = new PlainCache;
        ICache c = mc.bind (new CacheLoader);

        c.get ("abcde");
}


/*******************************************************************************


*******************************************************************************/

void testRemoteCacheLoader()
{
        class RemoteCacheLoader : Payload, IRemoteCacheLoader
        {
                bool test (IPayload p)
                {
                        return false;
                }            
                
                IPayload load (char[] key, long time)
                {
                        printf ("loading remote cache instance\n");
                        return new NullMessage;
                }    

                /**********************************************************************
                
                        Recover the timestamp from the provided reader

                **********************************************************************/

                void read (IReader reader)
                {
                        super.read (reader);
                }

		/**********************************************************************

			Emit our timestamp to the provided writer

		**********************************************************************/

		void write (IWriter writer)
		{
			super.write (writer);
		}

		/***********************************************************************

			Create a new instance of a payload, and populate it via
			read() using the specified reader

		***********************************************************************/

		Object create (IReader reader)
		{
			Payload r = new RemoteCacheLoader;
			r.read (reader);
			return r;
		}

		/**********************************************************************

			Return the guid for this payload. This should be unique
			per payload class, if said class is used in conjunction
			with the clustering facilities. Inspected by the Pickle
			utilitiy classes.

		**********************************************************************/

		char[] getGuid ()
		{
			return this.classinfo.name;
		}


                uint pause (uint wait)
                {
                        return 100_000;
                }
	}

        Cluster cluster = new Cluster (logger, 4567);
        NetworkCache nc = new NetworkCache (cluster, "my.channel");
        IRemoteCacheLoader cl = new RemoteCacheLoader;
        ICache c = nc.bind (cl);

        c.get ("abcde");
}


/*******************************************************************************


*******************************************************************************/

void testDisplay ()
{
        class Encoder : IStringEncoder
        {
                void char8 (void* src, uint count)
                {
                        printf ("%.*s", count, src);
                }

                void char16 (void* src, uint count)
                {
                        assert (0);
                }

                void char32 (void* src, uint count)
                {
                        assert (0);
                }
        }

        IWriter w = new DisplayWriter (Stdio.stdout.createBuffer());
        w.setStringEncoder (new Encoder);

        ubyte b = 129;
        uint ui = uint.max;
        int i = -3;
        
        w.put (1);
        w.put (' ');
        w.put (1.12);

        w.put(b).put(' ')
         .put(ui).put(' ')
         .put(i).put(' ')
         .put("hello")
         .cr()
         .flush();
}


/*******************************************************************************


*******************************************************************************/

void testTextWriter ()
{
        class Encoder1 : IStringEncoder
        {
                void char8 (void* src, uint count)
                {
                        printf ("%.*s", count, src);
                }

                void char16 (void* src, uint count)
                {
                        assert (0);
                }

                void char32 (void* src, uint count)
                {
                        assert (0);
                }
        }

        IWriter w = new TextWriter (Stdout.getBuffer(), ", ");
        w.setStringEncoder (new Encoder1);

        ubyte b = 129;
        uint ui = uint.max;
        int i = -3;

        w.put(b)
         .put(ui)
         .put(i)
         .put("hello")
         .cr();
        w.put("test");
        w.flush();
}


/*******************************************************************************


*******************************************************************************/

int main(char[][] args)
{   
        BasicConfigurator.configure ();
        //PropertyConfigurator.configure ("log.properties");
        logger = Logger.getLogger ("mango.unittest");
        //logger.setLevel (logger.Level.Error);
        //logger.addAppender (new FileAppender ("log.txt", new DateLayout));
        //logger.addAppender (new StdioAppender (new SimpleTimerLayout));

        try {
            //testDisplay();
            //testTextWriter();
            //testHashMap();
            //testRemoteCacheLoader();
            //testCacheLoader();
            //testProtocol();
            //testClusterMessage();
            //testRollingFileLog();
            //testClusterCache();
            //testClusterQueue();
            //testClusterServer();
            //testInvalidatee();
            //testMulticast();

            //testAppend();
            //testFile4();
            //testClassSerialization();
            //testFormat();
            //testLayout();
            //testFileSize();
            //testFilePath();
            //testFileList();
            //testScanFiles();
            //testFile1();
            //testFile5();
            //testToken1();
            //testToken2();
            //testColumn();
            //testSocket();
            //testStdin();
            //testBuffer();
            //testToken3();
            //testFileRegex();
            //testLineRegex();
            //testSocketRegex();
            //testRandomAccess();
            //testCompositeIO();
            //testDirectIO();
            testHttpServer();
            //testServletEngine();
            //testUri();
            //testHttpClient();
            //testHttpClient2();
            //testHttpClient3();
            //testQueuedCache();
            //testFileBucket();
            //testScanFiles();
            //testVirtualCache();
            //testQueuedCache();
            //testPlainCache();
            //testMappedFile();
            logger.info ("Done");
            } catch (Object x)
                    {
                    logger.fatal (x.toString);
                    }

        return 0;
}

