/*******************************************************************************

        @file Utf8.d
        
        Copyright (C) 2004 Kris Bell
        
        This software is provided 'as-is', without any express or implied
        warranty. In no event will the authors be held liable for damages
        of any kind arising from the use of this software.
        
        Permission is hereby granted to anyone to use this software for any 
        purpose, including commercial applications, and to alter it and/or 
        redistribute it freely, subject to the following restrictions:
        
        1. The origin of this software must not be misrepresented; you must 
           not claim that you wrote the original software. If you use this 
           software in a product, an acknowledgment within documentation of 
           said product would be appreciated but is not required.

        2. Altered source versions must be plainly marked as such, and must 
           not be misrepresented as being the original software.

        3. This notice may not be removed or altered from any distribution
           of the source.


                        ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


        @version        Initial version, August 2004      
        @author         Kris


*******************************************************************************/

module mango.xcode.Utf8;

private import mango.xcode.Transcoder;

/*******************************************************************************

        Fast UTF-8 to Unicode transcoder. These are really sensitive to
        small changes on 32bit x86 devices, because the register set of
        those devices is so small. Beware of subtle changes which might
        extend the execution-period by as much as 200% ...

        These routines were tuned on an Intel P3; other devices may work
        more efficiently with a slightly different approach, though this
        is likely to be reasonably optimal on AMD x86 CPUs also. These
        algorithms could benefit significantly from those extra AMD64 
        registers.

        Note that foreach can produce noticeable more efficient code than 
        equivalent for() loops, with either indices or pointers. The 0.98
        compiler version exhibited some rather erratic behavior over the
        course of testing: in particular, elapsed time of method execution
        is noticeably dependent upon its physical location within the file
        (or, more specifically, the enclosing class). Yes, it sure sounds 
        crazy that if you switch the order of encode() with decode() that 
	they will consistently execute slower than as currently arranged.        

        Finally, please note that these are between 5 and 30 times faster 
        than equivalent functions in the std.utf Phobos module (dependent
	upon the mix of char values). Those functions (strangely) often
	allocate memory on a character basis, so will become significantly 
	slower where there's heap-contention by multiple threads.
        
*******************************************************************************/

class TranscoderUtf8 : Transcoder
{
        /***********************************************************************

        ***********************************************************************/

        char[] toString ()
        {
                return "utf-8";
        }

        /***********************************************************************

                Encode UTF-8 up to a maximum of 4 bytes long (five & six
                byte variations are not supported). Throws an exception
                where the input dchar is greater than 0x10ffff.

        ***********************************************************************/

        char[] encode (dchar[] input, char[] output, out uint consumed)
        {
                char*   pOut;
                uint    eaten;
                int     space;
        
                pOut = output;
                space = output.length;
                foreach (dchar b; input)
                        {                        
                        if (b < 0x80)
                           {
                           if (space--)
                               *pOut++ = b;
                           else
                              break;
                           }
                        else
                           if (b < 0x8000)
                              {
                              if ((space -= 2) < 0)
                                   break;

                              pOut[0] = 0xc0 | ((b >> 6) & 0x3f);
                              pOut[1] = 0x80 | (b & 0x3f);
                              pOut += 2;
                              }
                           else
                              if (b < 0x80000)
                                 {
                                 if ((space -= 3) < 0)
                                      break;

                                 pOut[0] = 0xe0 | ((b >> 12) & 0x3f);
                                 pOut[1] = 0x80 | ((b >> 6)  & 0x3f);
                                 pOut[2] = 0x80 | (b & 0x3f);
                                 pOut += 3;
                                 }
                              else
                                 if (b < 0x110000)
                                    {
                                    if ((space -= 4) < 0)
                                         break;

                                    pOut[0] = 0xf0 | ((b >> 18) & 0x3f);
                                    pOut[1] = 0x80 | ((b >> 12) & 0x3f);
                                    pOut[2] = 0x80 | ((b >> 6)  & 0x3f);
                                    pOut[3] = 0x80 | (b & 0x3f);
                                    pOut += 4;
                                    }
                                 else
                                    {
                                    fault ("invalid input while encoding");
                                    return null;
                                    }
                        ++eaten;
                        }
                
                consumed = eaten;
                return output [0..(pOut - &output[0])];
        }


        /***********************************************************************

                Decode UTF-8 produced by the above encode() method. This
                executes notably faster than the validating version.
        
        ***********************************************************************/

        dchar[] decode (char[] input, dchar[] output, out uint consumed)
        {
                uint    produced;
                uint    available;
                char*   pIn = input;

                available = input.length;
                foreach (inout dchar d; output)
                        {
                        if (! available--)
                              break;

                        dchar b = cast(dchar) *pIn;
                        if (b & 0x80)
                           {
                           if (b < 0xe0)
                              {
                              if (! available--)
                                    break;

                              b &= 0x1f;
                              b = (b << 6) | (*++pIn & 0x3f);
                              }
                           else
                              if (b < 0xf0)
                                 {
                                 if (available < 2)
                                     break;

                                 b &= 0x0f;
                                 b = (b << 6) | (pIn[1] & 0x3f);
                                 b = (b << 6) | (pIn[2] & 0x3f);
                                 available -= 2;
                                 pIn += 2;
                                 }
                              else
                                 {
                                 if (available < 3)
                                     break;

                                 b &= 0x07;
                                 b = (b << 6) | (pIn[1] & 0x3f);
                                 b = (b << 6) | (pIn[2] & 0x3f);
                                 b = (b << 6) | (pIn[3] & 0x3f);
                                 available -= 3;
                                 pIn += 3;
                                 }
                           }    
                                       
                        d = b;
                        ++pIn;
                        ++produced;
                        }

                consumed = pIn - &input[0];
                return output [0..produced];
        }
}



/*******************************************************************************

        Fast UTF-8 to Unicode transcoder. This adds some validation to 
        the decoding process.
        
*******************************************************************************/

class TranscoderUtf8Checked : TranscoderUtf8
{
        /***********************************************************************

                Check the incoming stream for validity while decoding. This
                does not perform the full range of validations, but catches
                a large percentage of content errors due to the block-nature
                of these methods. That is; those cases where an error is not 
                spotted will typically lead to an error within the following
                segment. Errors not explicitly looked for are those relating
                to the five & six byte utf-8 combinations.

                Currently executes at about 2x of the unchecked version.
                
        ***********************************************************************/

        dchar[] decode (char[] input, dchar[] output, out uint consumed)
        {
                uint    produced;
                uint    available;
                char*   pIn = input;

                available = input.length;
                foreach (inout dchar d; output)
                        {
                        if (! available)
                              break;
                        
                        dchar b = cast(dchar) *pIn;
                        if (b & 0x80)
                           {
                           uint len = b < 0xe0 ? 1 : (b < 0xf0 ? 2 : 3);
                           if (available > len)
                              {
                              available -= len;
                              b &= ((1 << (6 - len)) - 1);

                              do {
                                 char c = *++pIn;
                                 if ((c & 0xc0) != 0x80)
                                      fault ("invalid input sequence");                                 

                                 b = (b << 6) | (c & 0x3f);
                                 } while (--len);

                              if (!(b < 0xd800 || 
                                   (b > 0xdfff && b <= 0x10ffff && 
                                    b != 0xfffe && b != 0xffff)))
                                    fault ("invalid unicode character");                                 
                              }
                           else
                              break;
                           }

                        --available;                      
                        d = b;
                        ++pIn;
                        ++produced;
                        }

                consumed = pIn - &input[0];
                return output [0..produced];
        }
}
